from __future__ import annotations

import argparse
from pathlib import Path

from tqecd.construction import annotate_detectors_automatically
from typing_extensions import override

from tqec._cli.subcommands.base import TQECSubCommand
from tqec._cli.subcommands.dae2observables import save_correlation_surfaces_to
from tqec.compile.compile import compile_block_graph
from tqec.compile.convention import FIXED_BULK_CONVENTION
from tqec.computation.block_graph import BlockGraph


class Dae2CircuitsTQECSubCommand(TQECSubCommand):
    @staticmethod
    @override
    def add_subcommand(
        main_parser: argparse._SubParsersAction[argparse.ArgumentParser],
    ) -> None:
        parser: argparse.ArgumentParser = main_parser.add_parser(
            "dae2circuits",
            description=(
                "Convert a .dae file representing a logical computation "
                "into concrete stim circuits."
            ),
        )
        parser.add_argument(
            "dae_file",
            help="A valid .dae file representing a computation.",
            type=Path,
        )
        parser.add_argument(
            "--out-dir",
            help="Directory to save the generated stim circuits.",
            type=Path,
            required=True,
        )
        parser.add_argument(
            "-k",
            help="The scale factors applied to the circuits.",
            nargs="+",
            type=int,
            required=True,
        )
        parser.add_argument(
            "--obs-include",
            help=(
                "The observable indices to be included in the circuits. "
                "If not provided, all potential observables will be included."
            ),
            nargs="*",
            type=int,
        )
        parser.add_argument(
            "--add-detectors",
            help="Whether to add detectors to the circuits.",
            action="store_true",
        )
        # TODO: add noise models
        parser.set_defaults(func=Dae2CircuitsTQECSubCommand.execute)

    @staticmethod
    @override
    def execute(args: argparse.Namespace) -> None:
        dae_absolute_path: Path = args.dae_file.resolve()
        out_dir: Path = args.out_dir.resolve()
        if not out_dir.exists():
            out_dir.mkdir(parents=True)

        # Construct the block graph from the given .dae file
        block_graph = BlockGraph.from_dae_file(
            dae_absolute_path, graph_name=str(dae_absolute_path.name)
        )

        # Convert to ZX graph and find observables
        zx_graph = block_graph.to_zx_graph()
        # Save the plotted observables to a subdirectory
        observable_out_dir = out_dir / "observables"
        observable_out_dir.mkdir(exist_ok=True)
        correlation_surfaces = block_graph.find_correlation_surfaces()
        obs_indices: list[int] = args.obs_include
        if not obs_indices:
            obs_indices = list(range(len(correlation_surfaces)))
        if max(obs_indices) >= len(correlation_surfaces):
            raise ValueError(
                f"Found {len(correlation_surfaces)} observables, but requested "
                f"indices up to {max(obs_indices)}."
            )

        save_correlation_surfaces_to(
            zx_graph,
            observable_out_dir,
            [correlation_surfaces[i] for i in obs_indices],
        )

        # Compile the block graph and generate stim circuits
        circuits_out_dir = out_dir / "circuits"
        circuits_out_dir.mkdir(exist_ok=True)
        compiled_graph = compile_block_graph(
            block_graph,
            FIXED_BULK_CONVENTION,
            observables=[correlation_surfaces[i] for i in obs_indices],
        )
        ks: list[int] = args.k
        add_detectors: bool = args.add_detectors
        for k in ks:
            circuit = compiled_graph.generate_stim_circuit(k)
            if add_detectors:
                circuit = annotate_detectors_automatically(circuit)
            circuit.to_file(circuits_out_dir / f"{k=}.stim")
            print(f"Write circuit to {circuits_out_dir}.")



from __future__ import annotations

from enum import Enum, auto


class PlaquetteOrientation(Enum):
    RIGHT = auto()
    LEFT = auto()
    DOWN = auto()
    UP = auto()

    def to_plaquette_side(self) -> PlaquetteSide:
        """Return the plaquette side corresponding to ``self``.

        If ``self == PlaquetteOrientation.RIGHT`` then
        ``self.to_plaquette_side() == PlaquetteSide.LEFT`` because a plaquette oriented to the right
        has 2 data-qubit on its left side.

        """
        if self == PlaquetteOrientation.RIGHT:
            return PlaquetteSide.LEFT
        elif self == PlaquetteOrientation.LEFT:
            return PlaquetteSide.RIGHT
        elif self == PlaquetteOrientation.DOWN:
            return PlaquetteSide.UP
        else:  # if self == PlaquetteOrientation.UP:
            return PlaquetteSide.DOWN


class PlaquetteSide(Enum):
    RIGHT = auto()
    LEFT = auto()
    DOWN = auto()
    UP = auto()

    def opposite(self) -> PlaquetteSide:
        """Return the opposite side."""
        if self == PlaquetteSide.RIGHT:
            return PlaquetteSide.LEFT
        elif self == PlaquetteSide.LEFT:
            return PlaquetteSide.RIGHT
        elif self == PlaquetteSide.DOWN:
            return PlaquetteSide.UP
        else:
            return PlaquetteSide.DOWN



from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING, Any

from tqec.circuit.qubit import GridQubit
from tqec.plaquette.enums import PlaquetteOrientation
from tqec.plaquette.rpng.rpng import RPNG, PauliBasis, RPNGDescription

if TYPE_CHECKING:
    from tqec.visualisation.computation.plaquette.base import SVGPlaquetteDrawer


@dataclass
class DrawPolygon:
    qubits_by_basis: dict[PauliBasis, list[GridQubit]] | PauliBasis

    def to_json(self) -> dict[str, Any]:  # pragma: no cover
        """Serialize ``self`` as a JSON-like dictionary."""
        if isinstance(self.qubits_by_basis, PauliBasis):
            return {"basis": self.qubits_by_basis.value}
        return {
            str(basis): [qubit.to_dict() for qubit in qubits]
            for basis, qubits in self.qubits_by_basis.items()
        }

    @staticmethod
    def from_json(data: dict[str, Any]) -> DrawPolygon:  # pragma: no cover
        """Deserialize ``self`` from a JSON-like dictionary."""
        if "basis" in data:
            return DrawPolygon(PauliBasis(data["basis"]))
        return DrawPolygon(
            {
                PauliBasis(basis): [GridQubit.from_dict(q) for q in qubits]
                for basis, qubits in data.items()
            }
        )


@dataclass(frozen=True)
class PlaquetteDebugInformation:
    rpng: RPNGDescription | None = None
    draw_polygons: DrawPolygon | None = None
    drawer: SVGPlaquetteDrawer | None = None

    def get_polygons(self) -> dict[PauliBasis, list[GridQubit]] | PauliBasis | None:
        """Getter method to get the polygon information for Crumble outputs.

        A polygon is one of Crumble annotation that allows to plot coloured polygons when exploring
        a quantum circuit with Crumble. That is useful to annotate stabilizers and their basis.

        Returns:
            a mapping from measurement basis to the data-qubits that are measured in each basis. If
            the debug information does not exists, returns ``None``. If all the data-qubits are
            measured in the same basis, returns that basis as a simplification.

        """
        if self.draw_polygons is not None:
            return self.draw_polygons.qubits_by_basis
        if self.rpng is not None:
            bases = {rpng.p for rpng in self.rpng.corners if rpng.p is not None}
            if len(bases) == 1:
                return bases.pop()
        return None

    def with_data_qubits_removed(self, removed_data_qubits: list[int]) -> PlaquetteDebugInformation:
        """Return a copy of ``self`` without any information on ``removed_data_qubits``."""
        if self.rpng is None:
            return self
        corners: list[RPNG] = list(self.rpng.corners)
        empty_rpng = RPNG.from_string("----")
        return PlaquetteDebugInformation(
            RPNGDescription(
                (
                    corners[0] if 0 not in removed_data_qubits else empty_rpng,
                    corners[1] if 1 not in removed_data_qubits else empty_rpng,
                    corners[2] if 2 not in removed_data_qubits else empty_rpng,
                    corners[3] if 3 not in removed_data_qubits else empty_rpng,
                ),
                self.rpng.ancilla,
            )
        )

    def project_on_boundary(
        self, projected_orientation: PlaquetteOrientation
    ) -> PlaquetteDebugInformation:
        """Project the debug information on the provided ``projected_orientation``."""
        if self.rpng is None:
            return self
        corners: list[RPNG] = list(self.rpng.corners)
        empty_rpng = RPNG.from_string("----")
        match projected_orientation:
            case PlaquetteOrientation.UP:
                corners[0] = corners[1] = empty_rpng
            case PlaquetteOrientation.DOWN:
                corners[2] = corners[3] = empty_rpng
            case PlaquetteOrientation.LEFT:
                corners[1] = corners[3] = empty_rpng
            case PlaquetteOrientation.RIGHT:
                corners[0] = corners[2] = empty_rpng
        return PlaquetteDebugInformation(
            RPNGDescription((corners[0], corners[1], corners[2], corners[3]), self.rpng.ancilla),
            self.draw_polygons,
        )

    def to_dict(self) -> dict[str, Any]:
        """Return a dictionary representing ``self``."""
        return {
            "rpng": self.rpng.to_dict() if self.rpng is not None else None,
            "draw_polygons": (
                self.draw_polygons.to_json() if self.draw_polygons is not None else None
            ),
        }

    @staticmethod
    def from_dict(data: dict[str, Any]) -> PlaquetteDebugInformation:
        """Initialise a :class:`.DebugInformation` instance from a dictionary."""
        return PlaquetteDebugInformation(
            (RPNGDescription.from_dict(data["rpng"]) if data["rpng"] is not None else None),
            (
                DrawPolygon.from_json(data["draw_polygons"])
                if data["draw_polygons"] is not None
                else None
            ),
        )

    def get_svg_drawer(self) -> SVGPlaquetteDrawer:  # pragma: no cover
        """Get a drawer to draw the plaquette associated to ``self``."""
        # Warning explicitly disabled because this intended and the only way to
        # avoid the costly svg import.
        from tqec.visualisation.computation.plaquette.base import (  # noqa: PLC0415
            EmptySVGPlaquetteDrawer,
        )
        from tqec.visualisation.computation.plaquette.rpng import (  # noqa: PLC0415
            RPNGPlaquetteDrawer,
        )

        if self.drawer is not None:
            return self.drawer
        if self.rpng is not None:
            return RPNGPlaquetteDrawer(self.rpng)
        return EmptySVGPlaquetteDrawer()



"""Define the face geometries used in COLLADA models."""

from __future__ import annotations

from dataclasses import astuple, dataclass, field
from typing import TYPE_CHECKING

import numpy as np
import numpy.typing as npt

from tqec.computation.cube import YHalfCube, ZXCube
from tqec.computation.pipe import PipeKind
from tqec.interop.color import TQECColor
from tqec.utils.enums import Basis
from tqec.utils.position import Direction3D, FloatPosition3D, SignedDirection3D

if TYPE_CHECKING:
    from tqec.computation.block_graph import BlockKind


@dataclass(frozen=True)
class Face:
    """A rectangle face in the 3d space.

    The axis of (width, height, normal_direction) is by the right-hand rule, i.e.
    (X, Y, Z) or (Y, Z, X) or (Z, X, Y).

    Attributes:
        color: The color of the face.
        width: The width of the face.
        height: The height of the face.
        normal_direction: The normal direction of the face, which is the direction
            of the axis that the face is perpendicular to.
        position: The position of the face in the 3D space.

    """

    color: TQECColor
    width: float
    height: float
    normal_direction: SignedDirection3D
    position: FloatPosition3D = field(default=FloatPosition3D(0.0, 0.0, 0.0))

    @staticmethod
    def get_triangle_indices() -> npt.NDArray[np.int_]:
        return np.array([0, 0, 2, 2, 1, 1, 2, 2, 0, 0, 3, 3], dtype=np.int_)

    def get_vertices(self) -> npt.NDArray[np.float64]:
        a3 = self.normal_direction.direction.value
        a1, a2 = (a3 + 1) % 3, (a3 + 2) % 3
        # rectangle vertices
        vertices_position = np.zeros((4, 3), dtype=np.float64)

        # Define rectangle vertices in the plane defined by axes a1 and a2
        vertices_position[:, a1] = [0, self.width, self.width, 0]
        vertices_position[:, a2] = [0, 0, self.height, self.height]
        # translate the rectangle
        vertices_position += np.asarray(astuple(self.position), dtype=np.float64)
        return vertices_position.flatten()

    def get_normal_vectors(self) -> npt.NDArray[np.float64]:
        normal = np.zeros(3, dtype=np.float64)
        ax3_index = self.normal_direction.direction.value
        normal[ax3_index] = 1 if self.normal_direction.towards_positive else -1
        return np.tile(normal, 4)

    def shift_by(self, dx: float, dy: float, dz: float) -> Face:
        return Face(
            self.color,
            self.width,
            self.height,
            self.normal_direction,
            self.position.shift_by(dx, dy, dz),
        )

    def with_negated_normal_direction(self) -> Face:
        return Face(
            self.color,
            self.width,
            self.height,
            -self.normal_direction,
            self.position,
        )


class BlockGeometries:
    """A collection of geometries of blocks."""

    def __init__(self) -> None:
        self.geometries: dict[BlockKind, list[Face]] = {}
        # 6 x/z cube blocks
        self._load_zx_cube_geometries()
        # 1 y half cube block
        self._load_y_cube_geometry()
        # 6 pipe blocks without H
        self._load_pipe_without_hadamard_geometries()
        # 6 pipe blocks with H
        self._load_pipe_with_hadamard_geometries()

    def get_geometry(
        self,
        kind: BlockKind,
        pop_faces_at_directions: frozenset[SignedDirection3D] = frozenset(),
    ) -> list[Face]:
        """Get the geometry of a block kind, possibly with some faces popped out."""
        faces = self.geometries[kind]
        return [face for face in faces if face.normal_direction not in pop_faces_at_directions]

    def _load_zx_cube_geometries(self) -> None:
        """Geometries for zxx, xzx, xxz, xzz, zxz, zzx cubes."""
        width, height = 1.0, 1.0
        for kind in ZXCube.all_kinds():
            faces: list[Face] = []
            for direction in Direction3D.all_directions():
                basis = kind.get_basis_along(direction)
                face = Face(
                    TQECColor(basis.value),
                    width,
                    height,
                    SignedDirection3D(direction, False),
                )
                faces.append(face)
                translation = [0.0, 0.0, 0.0]
                translation[direction.value] = 1.0
                faces.append(face.shift_by(*translation).with_negated_normal_direction())
            self.geometries[kind] = faces

    def _load_y_cube_geometry(self) -> None:
        """Geometry for the Y Half Cube."""
        faces: list[Face] = []
        for direction in Direction3D.all_directions():
            if direction == Direction3D.X:
                width, height = 1.0, 0.5
            elif direction == Direction3D.Y:
                width, height = 0.5, 1.0
            else:
                width, height = 1.0, 1.0
            face = Face(TQECColor.Y, width, height, SignedDirection3D(direction, False))
            faces.append(face)
            translation = [0.0, 0.0, 0.0]
            translation[direction.value] = 1.0 if direction != Direction3D.Z else 0.5
            faces.append(face.shift_by(*translation).with_negated_normal_direction())
        self.geometries[YHalfCube()] = faces

    def _load_pipe_without_hadamard_geometries(self) -> None:
        """Geometries for ozx, oxz, xoz, zox, xzo, zxo pipes."""
        for name in ["ozx", "oxz", "xoz", "zox", "xzo", "zxo"]:
            faces: list[Face] = []
            kind = PipeKind.from_str(name)
            for direction in Direction3D.all_directions():
                basis = kind.get_basis_along(direction)
                if basis is None:
                    continue
                if direction.value == (kind.direction.value - 1) % 3:
                    width, height = 2.0, 1.0
                else:
                    width, height = 1.0, 2.0
                face = Face(
                    TQECColor(basis.value),
                    width,
                    height,
                    SignedDirection3D(direction, False),
                )
                faces.append(face)
                translation = [0.0, 0.0, 0.0]
                translation[direction.value] = 1.0
                faces.append(face.shift_by(*translation).with_negated_normal_direction())
            self.geometries[kind] = faces

    def _load_pipe_with_hadamard_geometries(self) -> None:
        """Geometries for ozxh, oxzh, zoxh, xozh, zxoh, xzoh pipes."""

        def _get_face_position(shift_direction: Direction3D, shift: float) -> FloatPosition3D:
            tmp: list[float] = [0, 0, 0]
            tmp[shift_direction.value] = shift
            return FloatPosition3D(*tmp)

        for name in ["ozxh", "oxzh", "zoxh", "xozh", "zxoh", "xzoh"]:
            faces: list[Face] = []
            kind = PipeKind.from_str(name)
            for direction in Direction3D.all_directions():
                basis = kind.get_basis_along(direction)
                if basis is None:
                    continue
                if direction.value == (kind.direction.value - 1) % 3:
                    w1, h1 = 0.9, 1.0
                    w2, h2 = 0.2, 1.0
                    w3, h3 = 0.9, 1.0
                else:
                    w1, h1 = 1.0, 0.9
                    w2, h2 = 1.0, 0.2
                    w3, h3 = 1.0, 0.9
                normal_direction = SignedDirection3D(direction, False)
                face1 = Face(TQECColor(basis.value), w1, h1, normal_direction)
                face2 = Face(
                    TQECColor.H,
                    w2,
                    h2,
                    normal_direction,
                    _get_face_position(kind.direction, 0.9),
                )
                face3 = Face(
                    face1.color.with_zx_flipped(),
                    w3,
                    h3,
                    normal_direction,
                    _get_face_position(kind.direction, 1.1),
                )
                faces += [face1, face2, face3]
                translation = [0, 0, 0]
                translation[direction.value] = 1
                faces.extend(
                    face.shift_by(*translation).with_negated_normal_direction()
                    for face in [face1, face2, face3]
                )
            self.geometries[kind] = faces

    def _load_correlation_surface_geometries(self) -> None:
        """Geometries for horizontal, vertical and inclined surfaces."""
        # horizontal surface
        self.geometries  # pragma: no cover


def get_correlation_surface_geometry(basis: Basis) -> Face:
    """Get the face representing the correlation surface."""
    normal_direction = SignedDirection3D(Direction3D.Z, True)
    return Face(TQECColor(basis.value + "_CORRELATION"), 1.0, 1.0, normal_direction)



import functools
from collections.abc import Callable

from typing_extensions import override

from tqec.compile.blocks.block import Block
from tqec.compile.blocks.layers.atomic.base import BaseLayer
from tqec.compile.blocks.layers.atomic.plaquettes import PlaquetteLayer
from tqec.compile.blocks.layers.composed.base import BaseComposedLayer
from tqec.compile.blocks.layers.composed.repeated import RepeatedLayer
from tqec.compile.specs.base import (
    CubeBuilder,
    CubeSpec,
    PipeBuilder,
    PipeSpec,
)
from tqec.compile.specs.enums import SpatialArms
from tqec.compile.specs.library.generators.fixed_bulk import (
    FixedBulkConventionGenerator,
)
from tqec.computation.cube import Port, YHalfCube, ZXCube
from tqec.plaquette.compilation.base import IdentityPlaquetteCompiler, PlaquetteCompiler
from tqec.plaquette.plaquette import Plaquettes
from tqec.plaquette.rpng.translators.base import RPNGTranslator
from tqec.plaquette.rpng.translators.default import DefaultRPNGTranslator
from tqec.templates.base import RectangularTemplate
from tqec.utils.enums import Basis, Orientation
from tqec.utils.exceptions import TQECError
from tqec.utils.position import Direction3D
from tqec.utils.scale import LinearFunction


class FixedBulkCubeBuilder(CubeBuilder):
    def __init__(
        self,
        compiler: PlaquetteCompiler,
        translator: RPNGTranslator = DefaultRPNGTranslator(),
    ) -> None:
        """Implement the :class:`.CubeBuilder` interface for the fixed bulk convention.

        This class provides an implementation following the fixed-bulk convention. This convention
        consists in the fact that the top-left most plaquette in the bulk always measures a known-
        basis stabilizer (Z-basis for this class).

        """
        self._generator = FixedBulkConventionGenerator(translator, compiler)

    def _get_template_and_plaquettes(
        self, spec: CubeSpec
    ) -> tuple[RectangularTemplate, tuple[Plaquettes, Plaquettes, Plaquettes]]:
        """Get the template and plaquettes corresponding to the provided ``spec``.

        Args:
            spec: specification of the cube we want to implement.

        Returns:
            the template and list of 3 mappings from plaquette indices to RPNG
            descriptions that are needed to implement the cube corresponding to
            the provided ``spec``.

        """
        assert isinstance(spec.kind, ZXCube)
        x, _, z = spec.kind.as_tuple()
        if not spec.is_spatial:
            orientation = Orientation.HORIZONTAL if x == Basis.Z else Orientation.VERTICAL
            return self._generator.get_memory_qubit_raw_template(), (
                self._generator.get_memory_qubit_plaquettes(orientation, z, None),
                self._generator.get_memory_qubit_plaquettes(orientation, None, None),
                self._generator.get_memory_qubit_plaquettes(orientation, None, z),
            )
        # else:
        return self._generator.get_spatial_cube_qubit_raw_template(), (
            self._generator.get_spatial_cube_qubit_plaquettes(x, spec.spatial_arms, z, None),
            self._generator.get_spatial_cube_qubit_plaquettes(x, spec.spatial_arms, None, None),
            self._generator.get_spatial_cube_qubit_plaquettes(x, spec.spatial_arms, None, z),
        )

    @override
    def __call__(self, spec: CubeSpec, block_temporal_height: LinearFunction) -> Block:
        return self._call_impl(spec, block_temporal_height)

    @functools.cache
    def _call_impl(self, spec: CubeSpec, block_temporal_height: LinearFunction) -> Block:
        kind = spec.kind
        if isinstance(kind, Port):
            raise TQECError("Cannot build a block for a Port.")
        elif isinstance(kind, YHalfCube):
            raise NotImplementedError("Y cube is not implemented.")
        # else
        template, (init, repeat, measure) = self._get_template_and_plaquettes(spec)
        layers: list[BaseLayer | BaseComposedLayer] = [
            PlaquetteLayer(template, init),
            RepeatedLayer(PlaquetteLayer(template, repeat), repetitions=block_temporal_height),
            PlaquetteLayer(template, measure),
        ]
        return Block(layers)


class FixedBulkPipeBuilder(PipeBuilder):
    def __init__(
        self,
        compiler: PlaquetteCompiler,
        translator: RPNGTranslator = DefaultRPNGTranslator(),
    ) -> None:
        """Implement the :class:`.PipeBuilder` interface for the fixed bulk convention.

        This class provides an implementation following the fixed-bulk convention. This convention
        consists in the fact that the top-left most plaquette in the bulk always measures a known-
        parity stabilizer (Z-basis for this class).

        """
        self._generator = FixedBulkConventionGenerator(translator, compiler)

    @override
    def __call__(self, spec: PipeSpec, block_temporal_height: LinearFunction) -> Block:
        return self._call_impl(spec, block_temporal_height)

    @functools.cache
    def _call_impl(self, spec: PipeSpec, block_temporal_height: LinearFunction) -> Block:
        if spec.pipe_kind.is_temporal:
            return self._get_temporal_pipe_block(spec)
        return self._get_spatial_pipe_block(spec, block_temporal_height)

    #######################
    #    TEMPORAL PIPE    #
    #######################

    def _get_temporal_pipe_block(self, spec: PipeSpec) -> Block:
        """Return the block to implement a temporal pipe based on the provided ``spec``.

        Args:
            spec: description of the pipe that should be implemented by this
                method. Should be a temporal pipe.

        Raises:
            AssertionError: if ``spec`` does not represent a temporal junction.

        Returns:
            the block to implement a temporal pipe based on the
        provided ``spec``..

        """
        assert spec.pipe_kind.is_temporal
        if spec.pipe_kind.has_hadamard:
            return self._get_temporal_hadamard_pipe_block(spec)
        # Else, it is a regular temporal junction
        return self._get_temporal_non_hadamard_pipe_block(spec)

    def _get_temporal_non_hadamard_pipe_block(self, spec: PipeSpec) -> Block:
        """Return the block to implement a regular temporal junction without Hadamard transition.

        Args:
            spec: description of the pipe that should be implemented by this
                method. Should be a regular (i.e., non-Hadamard) temporal pipe.

        Raises:
            AssertionError: if the provided ``pipe`` is not a temporal pipe, or
                if it contains a Hadamard transition.

        Returns:
            the block to implement the provided
            ``spec``.

        """
        assert spec.pipe_kind.is_temporal
        assert not spec.pipe_kind.has_hadamard

        z_observable_orientation = (
            Orientation.HORIZONTAL if spec.pipe_kind.x == Basis.Z else Orientation.VERTICAL
        )
        memory_plaquettes = self._generator.get_memory_qubit_plaquettes(
            z_observable_orientation, None, None
        )
        template = self._generator.get_memory_qubit_raw_template()
        return Block(
            [
                PlaquetteLayer(template, memory_plaquettes)
                for _ in range(3 if spec.at_temporal_hadamard_layer else 2)
            ]
        )

    def _get_temporal_hadamard_pipe_block(self, spec: PipeSpec) -> Block:
        """Return the block to implement a temporal Hadamard pipe.

        Note:
            This method performs the realignment and Hadamard transition at the
            end of the layer that appear first (i.e., temporally before the other,
            or in other words the one with a lower Z index).

        Args:
            spec: description of the pipe that should be implemented by this
                method. Should be a Hadamard temporal pipe.

        Raises:
            AssertionError: if the provided ``pipe`` is not a temporal pipe, or
                if it is not a Hadamard transition.

        Returns:
            the block to implement the provided ``spec``.

        """
        assert spec.pipe_kind.is_temporal
        assert spec.pipe_kind.has_hadamard

        z_observable_orientation = (
            Orientation.HORIZONTAL if spec.pipe_kind.x == Basis.Z else Orientation.VERTICAL
        )
        memory_plaquettes_before = self._generator.get_memory_qubit_plaquettes(
            z_observable_orientation, None, None
        )
        realignment_plaquettes = self._generator.get_temporal_hadamard_realignment_plaquettes(
            z_observable_orientation
        )
        memory_plaquettes_after = self._generator.get_memory_qubit_plaquettes(
            z_observable_orientation.flip(), None, None
        )
        template = self._generator.get_temporal_hadamard_raw_template()
        return Block(
            [
                PlaquetteLayer(template, memory_plaquettes_before),
                PlaquetteLayer(template, realignment_plaquettes),
                PlaquetteLayer(template, memory_plaquettes_after),
            ]
        )

    ##############################
    #       SPATIAL PIPE         #
    ##############################
    @staticmethod
    def _get_spatial_cube_arms(spec: PipeSpec) -> SpatialArms:
        """Return the arm(s) corresponding to the provided ``spec``.

        Args:
            spec: pipe specification to get the arm(s) from.

        Raises:
            TQECError: if the provided ``spec`` is not a spatial pipe.

        Returns:
            the :class:`~tqec.compile.specs.enums.SpatialArms` instance
            corresponding to the provided ``spec``. The returned flag contains
            either one or two flags. If two flags are returned, they should be
            on the same line (e.g., it cannot be ``SpatialArms.RIGHT | SpatialArms.UP``
            but can be ``SpatialArms.RIGHT | SpatialArms.LEFT``).

        """
        assert spec.pipe_kind.is_spatial
        # Check that we do have a spatial junction.
        assert any(spec.is_spatial for spec in spec.cube_specs)
        u, v = spec.cube_specs
        pipedir = spec.pipe_kind.direction
        arms = SpatialArms.NONE
        if u.is_spatial:
            arms |= SpatialArms.RIGHT if pipedir == Direction3D.X else SpatialArms.DOWN
        if v.is_spatial:
            arms |= SpatialArms.LEFT if pipedir == Direction3D.X else SpatialArms.UP
        return arms

    def _get_spatial_cube_pipe_block(
        self, spec: PipeSpec, block_temporal_height: LinearFunction
    ) -> Block:
        x, y, z = spec.pipe_kind.x, spec.pipe_kind.y, spec.pipe_kind.z
        assert x is not None or y is not None
        spatial_boundary_basis: Basis = x if x is not None else y  # type: ignore
        # Get the plaquette indices mappings
        arms = FixedBulkPipeBuilder._get_spatial_cube_arms(spec)
        pipe_template = self._generator.get_spatial_cube_arm_raw_template(arms)
        initialisation_plaquettes = self._generator.get_spatial_cube_arm_plaquettes(
            spatial_boundary_basis, arms, spec.cube_specs, z, None
        )
        temporal_bulk_plaquettes = self._generator.get_spatial_cube_arm_plaquettes(
            spatial_boundary_basis, arms, spec.cube_specs, None, None
        )
        measurement_plaquettes = self._generator.get_spatial_cube_arm_plaquettes(
            spatial_boundary_basis, arms, spec.cube_specs, None, z
        )
        return Block(
            [
                PlaquetteLayer(pipe_template, initialisation_plaquettes),
                RepeatedLayer(
                    PlaquetteLayer(pipe_template, temporal_bulk_plaquettes),
                    repetitions=block_temporal_height,
                ),
                PlaquetteLayer(pipe_template, measurement_plaquettes),
            ]
        )

    def _get_spatial_regular_pipe_template(self, spec: PipeSpec) -> RectangularTemplate:
        """Return the template needed to implement the pipe representing the provided ``spec``."""
        assert spec.pipe_kind.is_spatial
        match spec.pipe_kind.direction, spec.pipe_kind.has_hadamard:
            case Direction3D.X, False:
                return self._generator.get_memory_vertical_boundary_raw_template()
            case Direction3D.X, True:
                return self._generator.get_spatial_vertical_hadamard_raw_template()
            case Direction3D.Y, False:
                return self._generator.get_memory_horizontal_boundary_raw_template()
            case Direction3D.Y, True:
                return self._generator.get_spatial_horizontal_hadamard_raw_template()
            case _:
                raise TQECError("Spatial pipes cannot have a direction equal to Direction3D.Z.")

    def _get_spatial_regular_pipe_plaquettes_factory(
        self, spec: PipeSpec
    ) -> Callable[[Basis | None, Basis | None], Plaquettes]:
        assert spec.pipe_kind.is_spatial
        match spec.pipe_kind.direction, spec.pipe_kind.has_hadamard:
            case Direction3D.X, False:
                # Non-Hadamard pipe in the X direction.
                z_observable_orientation = (
                    Orientation.HORIZONTAL if spec.pipe_kind.y == Basis.X else Orientation.VERTICAL
                )
                return lambda r, m: self._generator.get_memory_vertical_boundary_plaquettes(
                    z_observable_orientation, r, m
                )
            case Direction3D.X, True:
                # Hadamard pipe in the X direction.
                top_left_basis = spec.pipe_kind.get_basis_along(Direction3D.Y, at_head=True)
                return lambda r, m: self._generator.get_spatial_vertical_hadamard_plaquettes(
                    top_left_basis == Basis.Z, r, m
                )
            case Direction3D.Y, False:
                # Non-Hadamard pipe in the Y direction.
                z_observable_orientation = (
                    Orientation.HORIZONTAL if spec.pipe_kind.x == Basis.Z else Orientation.VERTICAL
                )
                return lambda r, m: self._generator.get_memory_horizontal_boundary_plaquettes(
                    z_observable_orientation, r, m
                )

            case Direction3D.Y, True:
                # Hadamard pipe in the Y direction.
                top_left_basis = spec.pipe_kind.get_basis_along(Direction3D.X, at_head=True)
                return lambda r, m: self._generator.get_spatial_horizontal_hadamard_plaquettes(
                    top_left_basis == Basis.Z, r, m
                )
            case _:
                raise TQECError("Spatial pipes cannot have a direction equal to Direction3D.Z.")

    def _get_spatial_regular_pipe_block(
        self, spec: PipeSpec, block_temporal_height: LinearFunction
    ) -> Block:
        assert all(not spec.is_spatial for spec in spec.cube_specs)
        plaquettes_factory = self._get_spatial_regular_pipe_plaquettes_factory(spec)
        template = self._get_spatial_regular_pipe_template(spec)

        layers: list[BaseLayer | BaseComposedLayer] = [
            PlaquetteLayer(
                template,
                plaquettes_factory(spec.pipe_kind.z, None),
            ),
            RepeatedLayer(
                PlaquetteLayer(
                    template,
                    plaquettes_factory(None, None),
                ),
                repetitions=block_temporal_height,
            ),
            PlaquetteLayer(
                template,
                plaquettes_factory(None, spec.pipe_kind.z),
            ),
        ]
        return Block(layers)

    def _get_spatial_pipe_block(
        self, spec: PipeSpec, block_temporal_height: LinearFunction
    ) -> Block:
        """Return the block to implement a spatial pipe based on the provided ``spec``.

        Args:
            spec: description of the pipe that should be implemented by this method. Should be a
                spatial pipe.
            block_temporal_height: the number of rounds of stabilizer measurements
            (ignoring one layer for initialization and another for final measurement).

        Raises:
            AssertionError: if ``spec`` does not represent a spatial pipe.

        Returns:
            the block to implement a spatial pipe based on the provided ``spec``.

        """
        assert spec.pipe_kind.is_spatial
        cube_specs = spec.cube_specs
        if cube_specs[0].is_spatial or cube_specs[1].is_spatial:
            return self._get_spatial_cube_pipe_block(spec, block_temporal_height)
        return self._get_spatial_regular_pipe_block(spec, block_temporal_height)


FIXED_BULK_CUBE_BUILDER: CubeBuilder = FixedBulkCubeBuilder(IdentityPlaquetteCompiler)
FIXED_BULK_PIPE_BUILDER: PipeBuilder = FixedBulkPipeBuilder(IdentityPlaquetteCompiler)



from __future__ import annotations

from abc import ABC, abstractmethod
from collections.abc import Mapping
from typing import TYPE_CHECKING

from tqec.compile.blocks.enums import TemporalBlockBorder
from tqec.utils.scale import LinearFunction

if TYPE_CHECKING:
    # Avoid circular dependency just because of typing annotations.
    from tqec.compile.blocks.layers.atomic.base import BaseLayer
    from tqec.compile.blocks.layers.composed.base import BaseComposedLayer


class WithTemporalFootprint(ABC):
    """Base class providing the interface implemented by objects that have a temporal footprint."""

    @property
    @abstractmethod
    def scalable_timesteps(self) -> LinearFunction:
        """Return the number of timesteps needed to implement the object.

        Returns:
            the number of timesteps needed to implement the object as an
            exact expression that can then be used to compute the number of
            timesteps for any value of ``k``.

        """
        pass

    def timesteps(self, k: int) -> int:
        """Return the number of timesteps needed for the provided scaling parameter ``k``.

        Args:
            k: scaling parameter.

        Returns:
            the number of timesteps needed to implement the object for
            the provided scaling parameter ``k``.

        """
        return self.scalable_timesteps.integer_eval(k)  # pragma: no cover

    @abstractmethod
    def with_temporal_borders_replaced(
        self, border_replacements: Mapping[TemporalBlockBorder, BaseLayer | None]
    ) -> BaseLayer | BaseComposedLayer | None:
        """Return ``self`` with the provided temporal borders replaced.

        Args:
            border_replacements: a mapping from temporal borders to replace to their replacement. A
                value of ``None`` as a replacement means that the border is removed.

        Returns:
            a copy of ``self`` with the provided ``borders`` replaced, or ``None``
            if replacing the provided ``borders`` from ``self`` result in an
            empty temporal footprint.

        """
        pass

    @abstractmethod
    def get_temporal_layer_on_border(self, border: TemporalBlockBorder) -> BaseLayer:
        """Return the :class:`.BaseLayer` on the provided ``border``.

        This method can be used to get the temporal extremal layers of a temporal pipe to replace
        the linked cube temporal extremal layers.

        """
        pass

    @property
    @abstractmethod
    def scalable_num_moments(self) -> LinearFunction:
        """Return the number of moments needed to implement the object,.

        Returns:
            the number of moments needed to implement the object as an
            exact expression that can then be used to compute the number of
            moments for any value of ``k``.

        """
        pass

    def num_moments(self, k: int) -> int:
        """Return the number of moments needed for the provided scaling parameter ``k``.

        Args:
            k: scaling parameter.

        Returns:
            the number of moments needed to implement the object for
            the provided scaling parameter ``k``.

        """
        return self.scalable_num_moments.integer_eval(k)  # pragma: no cover



"""Helper functions for representing correlation surfaces in COLLADA model."""

import collada
import numpy as np
import numpy.typing as npt

from tqec.computation.block_graph import BlockGraph
from tqec.computation.correlation import CorrelationSurface, ZXEdge
from tqec.computation.cube import Cube, ZXCube
from tqec.interop.collada.read_write import _Transformation
from tqec.utils.enums import Basis
from tqec.utils.position import Direction3D, FloatPosition3D, Position3D

TransformationResult = tuple[Basis, _Transformation]


class CorrelationSurfaceTransformationHelper:
    def __init__(self, block_graph: BlockGraph, pipe_length: float):
        """Help computing transformations of correlation surfaces pieces in a :class:`.BlockGraph`.

        The correlation surface is decomposed into small pieces of surfaces that can be transformed
        from a single 1x1 square surface in the XY-plane. This class computes the transformations
        for each piece of the correlation surface.

        """
        self.block_graph = block_graph
        self.pipe_length = pipe_length
        self.position_map = block_graph.to_zx_graph().positions

    def get_transformations_for_correlation_surface(
        self,
        correlation_surface: CorrelationSurface,
    ) -> list[TransformationResult]:
        """Return the transformations representing each piece of the ``correlation_surface``."""
        transformations: list[TransformationResult] = []

        # Surfaces in the pipes
        for edge in correlation_surface.span:
            if edge.is_self_loop:
                continue
            transformations.extend(self._compute_pipe_transformations(edge))

        # Surfaces in the cubes
        for v in correlation_surface.span_vertices:
            cube = self._get_cube(v)
            # Do not add surfaces in ports or Y Half Cubes
            if cube.is_port or cube.is_y_cube:
                continue
            transformations.extend(
                self._compute_cube_transformations(
                    v,
                    correlation_surface.edges_at(v),
                    correlation_surface.bases_at(v),
                )
            )
        return transformations

    def _get_position(self, v: int) -> Position3D:
        return self.position_map[v]

    def _get_cube(self, v: int) -> Cube:
        return self.block_graph[self.position_map[v]]

    def _edge_direction(self, edge: ZXEdge) -> Direction3D:
        """Return the edge direction."""
        p1, p2 = self._get_position(edge.u.id), self._get_position(edge.v.id)
        if p1.x != p2.x:
            return Direction3D.X
        if p1.y != p2.y:
            return Direction3D.Y
        return Direction3D.Z

    def _surface_normal_direction(
        self,
        correlation_edge: ZXEdge,
    ) -> Direction3D:
        """Get the correlation surface normal direction in the pipe."""
        u, v = correlation_edge
        up, vp = self._get_position(u.id), self._get_position(v.id)
        pipe = self.block_graph.get_pipe(up, vp)
        correlation_basis = u.basis
        return next(
            d
            for d in Direction3D.all_directions()
            if pipe.kind.get_basis_along(d) == correlation_basis.flipped()
        )

    def _scale_position(self, pos: Position3D) -> FloatPosition3D:
        return FloatPosition3D(*(p * (1 + self.pipe_length) for p in pos.as_tuple()))

    def _compute_pipe_transformations(self, edge: ZXEdge) -> list[TransformationResult]:
        """Compute the surface transformations within a pipe.

        If the edge is a Hadamard edge, two surfaces with different basis are created.

        """
        transformations: list[TransformationResult] = []
        normal_direction = self._surface_normal_direction(edge)
        edge_direction = self._edge_direction(edge)

        # Compute the translation for the surface.
        base_position = self._get_position(edge.u.id)
        scaled_position = self._scale_position(base_position)
        surface_position = scaled_position.shift_in_direction(edge_direction, 1).shift_in_direction(
            normal_direction, 0.5
        )
        rotation = _rotation_to_plane(normal_direction)
        scale_factor = self.pipe_length / 2 if edge.has_hadamard else self.pipe_length
        scale_direction = edge_direction if edge_direction != Direction3D.Z else normal_direction
        scale = _get_scale(scale_direction, scale_factor)

        transformations.append(
            (
                edge.u.basis,
                _Transformation(
                    translation=surface_position.as_array(),
                    rotation=rotation,
                    scale=scale,
                ),
            )
        )
        if edge.has_hadamard:
            transformations.append(
                (
                    edge.v.basis,
                    _Transformation(
                        translation=surface_position.shift_in_direction(
                            edge_direction, self.pipe_length / 2
                        ).as_array(),
                        rotation=rotation,
                        scale=scale,
                    ),
                ),
            )
        return transformations

    def _compute_cube_transformations(
        self,
        v: int,
        correlation_edges: set[ZXEdge],
        surface_bases: set[Basis],
    ) -> list[TransformationResult]:
        """Compute the transformations for the surfaces in the cube."""
        cube = self._get_cube(v)
        kind = cube.kind
        assert isinstance(kind, ZXCube)
        scaled_pos = self._scale_position(cube.position)
        transformations: list[TransformationResult] = []

        # Surfaces with even parity constraint
        if kind.normal_basis in surface_bases:
            normal_basis_edges: set[ZXEdge] = set()
            for edge in correlation_edges:
                this_node = edge.u if edge.u.id == v else edge.v
                if this_node.basis == kind.normal_basis:
                    normal_basis_edges.add(edge)
            assert len(normal_basis_edges) in {2, 4}, "Even parity constraint violated"
            if len(normal_basis_edges) == 2:
                e1, e2 = sorted(normal_basis_edges)
                # passthrough
                if self._edge_direction(e1) == self._edge_direction(e2):
                    normal_direction = self._surface_normal_direction(e1)
                    translation = scaled_pos.shift_in_direction(normal_direction, 0.5)
                    transformations.append(
                        (
                            kind.normal_basis,
                            _Transformation(
                                translation=translation.as_array(),
                                rotation=_rotation_to_plane(normal_direction),
                                scale=np.ones(3, dtype=np.float32),
                            ),
                        )
                    )
                # turn at corner
                else:
                    transformations.extend(
                        self._compute_turn_transformation(scaled_pos, v, (e1, e2))
                    )
            else:
                e1, e2, e3, e4 = sorted(normal_basis_edges)
                transformations.extend(self._compute_turn_transformation(scaled_pos, v, (e1, e2)))
                transformations.extend(self._compute_turn_transformation(scaled_pos, v, (e3, e4)))

        # Surfaces that can broadcast to all the neighbors
        if len(surface_bases) == 2 or kind.normal_basis not in surface_bases:
            translation = scaled_pos.shift_in_direction(kind.normal_direction, 0.5)
            transformations.append(
                (
                    kind.normal_basis.flipped(),
                    _Transformation(
                        translation=translation.as_array(),
                        scale=np.ones(3, dtype=np.float32),
                        rotation=_rotation_to_plane(kind.normal_direction),
                    ),
                )
            )
        return transformations

    def _compute_turn_transformation(
        self,
        cube_pos: FloatPosition3D,
        v: int,
        turn_edges: tuple[ZXEdge, ZXEdge],
    ) -> list[TransformationResult]:
        """Compute the transformations for the surfaces in a L-shape turn.

        At turn, two surfaces in the same basis are created and form a 90 degree angle.

        """
        cube_kind = self._get_cube(v).kind
        assert isinstance(cube_kind, ZXCube)

        transformations = []
        turn_dirs = [self._edge_direction(e) for e in turn_edges]
        for i, e in enumerate(turn_edges):
            d, other_d = turn_dirs[i], turn_dirs[1 - i]
            # scale transformation is applied before rotation and the primitive
            # surface is in the XY-plane, so we also need to scale in the XY-plane.
            scale_direction = d if d != Direction3D.Z else other_d
            scale = _get_scale(scale_direction, 0.5)
            rotation = _rotation_to_plane(other_d)
            translation = cube_pos.shift_in_direction(other_d, 0.5)
            if v == e.u.id:
                translation = translation.shift_in_direction(d, 0.5)
            transformations.append(
                (
                    cube_kind.normal_basis,
                    _Transformation(
                        translation=translation.as_array(),
                        rotation=rotation,
                        scale=scale,
                    ),
                )
            )
        return transformations


def _rotation_to_plane(
    plane_normal_direction: Direction3D,
) -> npt.NDArray[np.float32]:
    """Rotate a surface in the XY-plane to the given plane."""
    if plane_normal_direction == Direction3D.Z:
        return _rotation_matrix(Direction3D.Z, 0.0)
    elif plane_normal_direction == Direction3D.X:
        return _rotation_matrix(Direction3D.Y, 90.0)
    else:
        return _rotation_matrix(Direction3D.X, 90.0)  # pragma: no cover


def _rotation_matrix(
    axis: Direction3D,
    angle: float = 90.0,
) -> npt.NDArray[np.float32]:
    if axis == Direction3D.Y:
        angle = -angle
    axis_vec = np.zeros(3, dtype=np.float32)
    axis_vec[axis.value] = 1.0
    return np.asarray(
        collada.scene.RotateTransform(axis_vec[0], axis_vec[1], axis_vec[2], angle).matrix[:3, :3],
        dtype=np.float32,
    )


def _get_scale(scale_direction: Direction3D, scale_factor: float) -> npt.NDArray[np.float32]:
    scale = np.ones(3, dtype=np.float32)
    scale[scale_direction.value] = scale_factor
    return scale



import functools
import operator
from collections.abc import Iterator

import stim
from typing_extensions import override

from tqec.circuit.moment import Moment
from tqec.circuit.schedule.circuit import ScheduledCircuit
from tqec.plaquette.compilation.passes.base import CompilationPass


def _with_targets_sorted(moments: Iterator[Moment]) -> list[Moment]:
    ret_moments: list[Moment] = []
    for moment in moments:
        circuit = stim.Circuit()
        for instruction in moment.instructions:
            target_groups = instruction.target_groups()
            target_groups.sort(key=lambda group: [trgt.value for trgt in group])
            targets: list[stim.GateTarget] = functools.reduce(operator.iadd, target_groups, [])
            circuit.append(instruction.name, targets, instruction.gate_args_copy())
        ret_moments.append(Moment(circuit, moment.qubits_indices, _avoid_checks=True))
    return ret_moments


class SortTargetsPass(CompilationPass):
    """Compilation pass sorting the targets of the provided quantum circuit instructions."""

    @override
    def run(self, circuit: ScheduledCircuit, check_all_flows: bool = False) -> ScheduledCircuit:
        modified_circuit = ScheduledCircuit(
            _with_targets_sorted(circuit.moments), circuit.schedule, circuit.qubit_map
        )
        if check_all_flows:
            self.check_flows(circuit, modified_circuit)  # pragma: no cover
        return modified_circuit



from .rpng import RG as RG
from .rpng import RPNG as RPNG
from .rpng import ExtendedBasis as ExtendedBasis
from .rpng import PauliBasis as PauliBasis
from .rpng import RPNGDescription as RPNGDescription



"""Defines :class:`.MeasurementRecordsMap` which is used to query measurement offsets in a circuit.

This module defines :class:`MeasurementRecordsMap`. This class implements the
necessary interface to register, modify and query measurement offsets from a
``stim.Circuit``.

"""

from __future__ import annotations

from collections.abc import Sequence
from dataclasses import dataclass, field

import numpy
import stim

from tqec.circuit.qubit import GridQubit
from tqec.circuit.qubit_map import QubitMap
from tqec.circuit.schedule import ScheduledCircuit
from tqec.utils.exceptions import TQECError
from tqec.utils.instructions import (
    is_multi_qubit_measurement_instruction,
    is_single_qubit_measurement_instruction,
)


@dataclass(frozen=True)
class MeasurementRecordsMap:
    """A mapping from measurements appearing in a circuit and their record offsets.

    This class stores record offsets which are, by essence, relative to a certain
    position in a circuit. This means that this class and the measurement offsets
    it stores are meaningless without knowledge about the circuit containing the
    represented measurements and the position(s) in the circuit at which the
    instance at hand is valid.

    Raises:
        TQECError: if at least one of the provided measurement record offsets
            is non-negative (``>=0``).
        TQECError: if, for any of the provided qubits, the provided offsets
            are not sorted.
        TQECError: if any measurement offset is duplicated.

    """

    mapping: dict[GridQubit, list[int]] = field(default_factory=dict)

    def __post_init__(self) -> None:
        all_measurement_records_indices: list[int] = []
        for qubit, measurement_record_offsets in self.mapping.items():
            # Check that the provided measurement record offsets are negative.
            nonnegative_offsets = [offset for offset in measurement_record_offsets if offset >= 0]
            if nonnegative_offsets:
                raise TQECError(
                    "Invalid mapping from qubit offsets to measurement record "
                    f"offsets. Found positive offsets ({nonnegative_offsets}) for "
                    f"qubit {qubit}."
                )
            # Check that measurement record offsets are sorted
            if measurement_record_offsets != sorted(measurement_record_offsets):
                raise TQECError(
                    "Got measurement record offsets that are not in sorted "
                    f"order: {measurement_record_offsets}. This is not supported."
                )
            all_measurement_records_indices.extend(measurement_record_offsets)
        # Check that a given measurement record offset only appears once.
        deduplicated_indices = numpy.unique(all_measurement_records_indices)
        if len(deduplicated_indices) != len(all_measurement_records_indices):
            raise TQECError(
                "At least one measurement record offset has been found "
                "twice in the provided offsets."
            )

    @staticmethod
    def from_scheduled_circuit(circuit: ScheduledCircuit) -> MeasurementRecordsMap:
        """Build a :class:`MeasurementRecordsMap` from a scheduled circuit.

        Args:
            circuit: circuit containing the measurements to map.

        Raises:
            TQECError: if the provided ``circuit`` contains a ``REPEAT`` block.
            TQECError: if the provided ``circuit`` contains an unsupported
                measurement instruction. Currently, this method supports all
                single-qubit measurement instructions (see the value of
                ``SINGLE_QUBIT_MEASUREMENT_INSTRUCTION_NAMES``).
            TQECError: if one of the measurement counted in
                ``circuit.num_measurements`` was not processed (this should never
                raise, but the post-condition is easy to check so this method
                performs the check just in case).

        Returns:
            a :class:`MeasurementRecordsMap` linking measurements in the provided
            ``circuit`` to their offset.

        """
        return MeasurementRecordsMap.from_circuit(
            circuit.get_circuit(include_qubit_coords=False), circuit.qubit_map
        )

    @staticmethod
    def from_circuit(
        circuit: stim.Circuit, qubit_map: QubitMap | None = None
    ) -> MeasurementRecordsMap:
        """Build a :class:`MeasurementRecordsMap` from a circuit.

        Args:
            circuit: circuit containing the measurements to map.
            qubit_map: qubit map of the provided circuit. If ``None``, the
                qubit map is computed from the provided circuit. Default to
                ``None``.

        Raises:
            TQECError: if the provided ``circuit`` contains a ``REPEAT`` block.
            TQECError: if the provided ``circuit`` contains an unsupported
                measurement instruction. Currently, this method supports all
                single-qubit measurement instructions (see the value of
                ``SINGLE_QUBIT_MEASUREMENT_INSTRUCTION_NAMES``).
            TQECError: if one of the measurement counted in
                ``circuit.num_measurements`` was not processed (this should never
                raise, but the post-condition is easy to check so this method
                performs the check just in case).

        Returns:
            a :class:`MeasurementRecordsMap` linking measurements in the provided
            ``circuit`` to their offset.

        """
        # We iterate the circuit in forward order, which means that the first
        # measurement we will encounter will have a record offset of
        # `-circuit.num_measurements`. Iterating in this order also allows to
        # ensure by construction that the measurement record offsets will be in
        # sorted order.
        current_measurement_record = -circuit.num_measurements
        if qubit_map is None:
            qubit_map = QubitMap.from_circuit(circuit)
        measurement_records: dict[GridQubit, list[int]] = {}
        for instruction in circuit:
            if isinstance(instruction, stim.CircuitRepeatBlock):
                raise TQECError("Found a REPEAT instruction. This is not supported for the moment.")
            if is_multi_qubit_measurement_instruction(instruction):
                raise TQECError(f"Found a non-supported measurement instruction: {instruction}")
            if is_single_qubit_measurement_instruction(instruction):
                for (qi,) in instruction.target_groups():
                    qubit = qubit_map.i2q[qi.value]
                    measurement_records.setdefault(qubit, []).append(current_measurement_record)
                    current_measurement_record += 1
        if current_measurement_record != 0:
            raise TQECError(
                "Failed post-condition check. Expected a final measurement record of "
                f"-1 but got {current_measurement_record - 1}. Did we miss a "
                f"measurement? Circuit:\n{circuit}"
            )
        return MeasurementRecordsMap(measurement_records)

    # Explicitly returns a Sequence to show that the returned value is read-only.
    def __getitem__(self, qubit: GridQubit) -> Sequence[int]:
        return self.mapping[qubit]

    def __contains__(self, qubit: GridQubit) -> bool:
        return qubit in self.mapping

    def with_added_measurements(
        self, mrecords_map: MeasurementRecordsMap, repetitions: int = 1
    ) -> MeasurementRecordsMap:
        """Build a new instance with measurements from ``self`` and ``mrecords_map``.

        Measurements from ``self`` appear before measurements from the provided ``mrecords_map``.

        Args:
            mrecords_map: records of measurements happening after the measurements
                represented by ``self``.
            repetitions: number of time the measurements from ``mrecords_map`` are
                repeated. Default to 1.

        Returns:
            a new instance containing valid offsets for each measurement in ``self``
            and ``mrecords_map``.

        """
        num_measurements_without_repetition = sum(
            len(offsets) for offsets in mrecords_map.mapping.values()
        )
        num_added_measurements = repetitions * num_measurements_without_repetition
        records = {
            q: [o - num_added_measurements for o in offsets] for q, offsets in self.mapping.items()
        }
        for q, offsets in mrecords_map.mapping.items():
            record = records.setdefault(q, [])
            for i in range(repetitions - 1, -1, -1):
                record.extend([o - i * num_measurements_without_repetition for o in offsets])
        return MeasurementRecordsMap(records)



from typing_extensions import override

from tqec.compile.blocks.layers.atomic.layout import LayoutLayer
from tqec.compile.tree.node import LayerNode, NodeWalker


class AnnotateCircuitOnLayerNode(NodeWalker):
    def __init__(self, k: int, reschedule_measurements: bool = True):
        """Walker annotating ``stim.Circuit`` instances implementing the node on each leaf node.

        Args:
            k: scaling factor.
            reschedule_measurements: whether to reschedule measurements in the generated circuits
                to be in the same moment. Since each plaquette may have its own measurement
                schedule, setting this may be necessary for hardware that requires
                measurements to be synchronous.

        """
        self._k = k
        self._reschedule_measurements = reschedule_measurements

    @override
    def visit_node(self, node: LayerNode) -> None:
        if not node.is_leaf:
            return
        assert isinstance(node._layer, LayoutLayer)
        node.set_circuit_annotation(
            self._k,
            node._layer.to_circuit(self._k, reschedule_measurements=self._reschedule_measurements),
        )



from __future__ import annotations

import argparse
from abc import ABC, abstractmethod


class TQECSubCommand(ABC):
    """Interface that should be implemented by subcommands of the tqec CLI."""

    @staticmethod
    @abstractmethod
    def add_subcommand(
        main_parser: argparse._SubParsersAction[argparse.ArgumentParser],
    ) -> None:
        """Create a CLI for the subcommand using the provided ``main_parser``."""
        pass

    @staticmethod
    @abstractmethod
    def execute(args: argparse.Namespace) -> None:
        """Execute the subcommand from the provided parsed arguments."""
        pass



from __future__ import annotations

import stim

from tqec.circuit.moment import Moment
from tqec.post_processing.remove import _remove_empty_moments_inline
from tqec.post_processing.utils.moment import (
    RepeatedMoments,
    collect_moments,
    iter_stim_circuit_by_moments,
    merge_moments,
)


def merge_adjacent_moments(circuit: stim.Circuit) -> stim.Circuit:
    """Merge adjacent moments that can be merged.

    This function merges moments from the provided ``circuit`` that can be
    merged together. It performs merging in the increasing time direction. It
    may re-organise the neighbourhood of ``REPEAT`` blocks in some cases, but
    the returned circuit should always be strictly equivalent to the provided
    one.

    Args:
        circuit: a circuit containing moments that may be mergeable.

    Returns:
        a new quantum circuit that contains strictly the same instructions as
        the provided ``circuit`` but that might re-order some of these operations
        to reduce the number of moments.

    """
    merged_moments: list[Moment | RepeatedMoments] = list(
        iter_stim_circuit_by_moments(circuit, collected_before_use=True)
    )
    _remove_empty_moments_inline(merged_moments)
    modification_performed: bool = True
    while modification_performed:
        modification_performed = False
        modification_performed |= _merge_internal_adjacent_moments_inline(merged_moments)
        modification_performed |= _merge_repeat_block_boundaries_inline(merged_moments)
    return collect_moments(merged_moments)


def _can_be_merged(lhs: Moment, rhs: Moment) -> bool:
    return not lhs.qubits_indices.intersection(rhs.qubits_indices)


def _merge_internal_adjacent_moments_inline(
    moments: list[Moment | RepeatedMoments],
) -> bool:
    """Merge adjacent moments without considering REPEAT block boundaries.

    Returns:
        ``True`` if the provided ``moments`` have been modified, else ``False``.

    """
    i: int = 1
    modification_performed: bool = False
    while i < len(moments):
        # Invariant of the loop: moments[i - 1] is not a REPEAT block.
        previous_moment, current_moment = moments[i - 1], moments[i]
        assert not isinstance(previous_moment, RepeatedMoments)

        if isinstance(current_moment, RepeatedMoments):
            # Just recurse in the REPEAT block but do not perform any specific
            # computation for its boundaries.
            modification_performed |= _merge_internal_adjacent_moments_inline(
                current_moment.moments
            )
            # We do not have to look at the moment just after the REPEAT block
            # because it will not be merged with the REPEAT block, so just skip
            # it (that also allows to respect the loop invariant).
            i += 2
            continue

        # Try to merge moments[i] into moments[i - 1]
        # If they can be merged, merge them. Else, just increase i and go to the
        # next moment.
        if _can_be_merged(previous_moment, current_moment):
            moments[i - 1] = merge_moments(previous_moment, current_moment)
            moments.pop(i)
            modification_performed = True
        else:
            i += 1
    return modification_performed


def _merge_repeat_block_boundaries_inline(
    moments: list[Moment | RepeatedMoments],
) -> bool:
    i: int = 0
    modification_performed: bool = False
    while i < len(moments):
        current_moment = moments[i]
        if not isinstance(current_moment, RepeatedMoments):
            i += 1
            continue
        # First recursively merge the potentially nested REPEAT blocks boundaries.
        modification_performed |= _merge_repeat_block_boundaries_inline(current_moment.moments)
        # Then merge the boundaries of the current REPEAT block if possible.
        start, *bulk, end = current_moment.moments
        if isinstance(start, RepeatedMoments) or isinstance(end, RepeatedMoments):
            # Not handled yet, just ignore.
            i += 1
            continue
        if not _can_be_merged(start, end) or current_moment.repetitions < 2:
            # Either START and END cannot be merged or merging them would lead to
            # an invalid REPEAT block (repetitions == 0), so do nothing.
            i += 1
            continue
        # Here, we know that start and start can be merged. We replace:
        #
        # REPEAT N {
        #     START
        #     BULK
        #     END
        # }
        #
        # by
        #
        # START
        # BULK
        # REPEAT N - 1 {
        #     END + START
        #     BULK
        # }
        # END
        # Note: modifying moments in-place.
        moments.pop(i)
        new_repeat_block = RepeatedMoments(
            current_moment.repetitions - 1, [merge_moments(end, start), *bulk]
        )
        inserted_moments = [start, *bulk, new_repeat_block, end]
        for m in range(len(inserted_moments)):
            moments.insert(i + m, inserted_moments[m])
        modification_performed = True
        # We can update i to the index of the REPEAT block, just in case more
        # merging can be done. Note that even if BULK contains REPEAT blocks,
        # they have been addressed in the beginning of this loop iteration, hence
        # we do not need to iterate again on BULK.
        i += len(bulk) + 1
    return modification_performed



import stim

from tqec.circuit.moment import Moment
from tqec.post_processing.utils.moment import RepeatedMoments


def remove_empty_moments(
    circuit: stim.Circuit,
    remove_leading_tick: bool = True,
    remove_trailing_tick: bool = True,
) -> stim.Circuit:
    """Remove empty moments in a circuit.

    This function removes empty moments (delimited by ``TICK`` instructions) in
    the provided circuit, returning a copy of the circuit without the empty
    moments.

    Note:
        This function follows ``stim`` convention on ``REPEAT`` blocks to:

        1. not include any ``TICK`` instruction just before the ``REPEAT`` block,
        2. include a ``TICK`` instruction at the beginning of the repeated block,
        3. not include a ``TICK`` instruction at the end of the repeated block.

        The instruction following a ``REPEAT`` block with ``stim gen`` is often
        not a ``TICK`` because only the last measurements on data-qubits are left
        to perform and so they do not overlap with the last measurements within
        the repeated block. For this reason, this function do not enforce
        anything about the instruction following a ``REPEAT`` block:

        - if it is not a ``TICK`` in the original circuit, no ``TICK`` is
          inserted,
        - if there are one or more consecutive ``TICK`` instructions, only one
          will be left (removing empty moments).

    Args:
        circuit: circuit to remove empty moments from.
        remove_leading_tick: whether to remove the first instruction of
            ``circuit`` if it is a ``TICK`` instruction. Note that this parameter
            is not recursively forwarded to repeated blocks, see the note.
        remove_trailing_tick: whether to remove the last instruction of
            ``circuit`` if it is a ``TICK`` instruction. Note that this parameter
            is not recursively forwarded to repeated blocks, see the note.

    Returns:
        A quantum circuit without any empty moment and with ``REPEAT`` blocks
        adhering to the conventions detailed in the note above.

    """
    ret = stim.Circuit()
    # Start with a virtual TICK if the user wants to remove empty moments at the
    # beginning of the provided circuit. Else, any other instruction that is not
    # a TICK would work and avoid the removal of the first instruction if it is
    # a TICK.
    previous_instruction = (
        stim.CircuitInstruction("TICK")
        if remove_leading_tick
        else stim.CircuitInstruction("SHIFT_COORDS", [], [0])
    )
    for inst in circuit:
        # If we have two (or more) consecutive TICK instructions, do not append
        # any except the first one.
        if inst.name == previous_instruction.name == "TICK":
            continue
        # If we have a REPEAT block, be careful about the conventions.
        if isinstance(inst, stim.CircuitRepeatBlock):
            # Check convention 1 about the instruction before the REPEAT instruction
            if ret[-1].name == "TICK":
                ret = ret[:-1]
            # Check conventions 2 and 3 about the repeated block.
            repeated_block = stim.Circuit("TICK") + remove_empty_moments(
                inst.body_copy(), remove_leading_tick=True, remove_trailing_tick=True
            )
            ret.append(stim.CircuitRepeatBlock(inst.repeat_count, repeated_block))
        else:
            ret.append(inst)
        previous_instruction = inst
    # If the last instruction is a TICK, that's an empty moment, remove it.
    if previous_instruction.name == "TICK" and remove_trailing_tick:
        return ret[:-1]
    return ret


def _remove_empty_moments_inline(moments: list[Moment | RepeatedMoments]) -> None:
    i = 0
    while i < len(moments):
        moment = moments[i]
        if isinstance(moment, Moment):
            if moment.is_empty:
                moments.pop(i)
            else:
                i += 1
        else:
            _remove_empty_moments_inline(moment.moments)
            if len(moment.moments) == 0:
                moments.pop(i)
            else:
                i += 1



from __future__ import annotations

from dataclasses import dataclass
from typing import Literal, Protocol

from tqec.compile.blocks.block import Block
from tqec.compile.specs.enums import SpatialArms
from tqec.computation.block_graph import BlockGraph
from tqec.computation.cube import Cube, CubeKind, ZXCube
from tqec.computation.pipe import PipeKind
from tqec.templates.base import RectangularTemplate
from tqec.utils.exceptions import TQECError
from tqec.utils.position import Direction3D
from tqec.utils.scale import LinearFunction


@dataclass(frozen=True)
class CubeSpec:
    """Specification of a cube in a block graph.

    The template of the `CompiledBlock` will be determined based on the specification.
    This class can be used as a key to look up the corresponding `CompiledBlock` before
    applying the substitution rules.

    Attributes:
        cube_kind: The kind of the cube.
        spatial_arms: Flag indicating the spatial directions the cube connects to the
            adjacent cubes. This is useful for spatial cubes (XXZ and ZZX) where
            the arms can determine the template used to implement the cube.
        has_spatial_up_or_down_pipe_in_timeslice: a flag indicating if a spatial
            pipe at the top or bottom of a spatial cube is executed on the same
            timeslice as this cube. This information is needed for the fixed
            boundary convention.

    """

    kind: CubeKind
    spatial_arms: SpatialArms = SpatialArms.NONE
    has_spatial_up_or_down_pipe_in_timeslice: bool = False

    def __post_init__(self) -> None:
        if self.spatial_arms != SpatialArms.NONE:
            if not self.is_spatial:
                raise TQECError(
                    "The `spatial_arms` attribute should be `SpatialArms.NONE` "
                    "for non-spatial cubes."
                )

    @property
    def is_spatial(self) -> bool:
        """Return ``True`` if ``self`` represents a spatial cube."""
        return isinstance(self.kind, ZXCube) and self.kind.is_spatial

    @staticmethod
    def from_cube(
        cube: Cube,
        graph: BlockGraph,
        spatial_up_or_down_pipes_slices: frozenset[int] = frozenset(),
    ) -> CubeSpec:
        """Return the cube spec from a cube in a block graph."""
        has_spatial_up_or_down_pipe_in_timeslice = (
            cube.position.z in spatial_up_or_down_pipes_slices
        )
        if not cube.is_spatial:
            return CubeSpec(
                cube.kind,
                has_spatial_up_or_down_pipe_in_timeslice=has_spatial_up_or_down_pipe_in_timeslice,
            )
        spatial_arms = SpatialArms.from_cube_in_graph(cube, graph)
        return CubeSpec(cube.kind, spatial_arms, has_spatial_up_or_down_pipe_in_timeslice)

    @property
    def pipe_dimensions(self) -> frozenset[Literal[Direction3D.X, Direction3D.Y]]:
        """Return the dimension(s) in which ``self`` has at least one pipe."""
        dimensions: list[Literal[Direction3D.X, Direction3D.Y]] = []
        if SpatialArms.LEFT in self.spatial_arms or SpatialArms.RIGHT in self.spatial_arms:
            dimensions.append(Direction3D.X)
        if SpatialArms.UP in self.spatial_arms or SpatialArms.DOWN in self.spatial_arms:
            dimensions.append(Direction3D.Y)
        return frozenset(dimensions)

    @property
    def has_spatial_pipe_in_both_dimensions(self) -> bool:
        """Return ``True`` if the provided spec has a pipe in each of the two spatial dimensions."""
        return self.spatial_arms.has_spatial_arm_in_both_dimensions


class CubeBuilder(Protocol):
    """Protocol for building a `Block` based on a `CubeSpec`."""

    def __call__(self, spec: CubeSpec, block_temporal_height: LinearFunction) -> Block:
        """Build a ``Block`` instance from a ``CubeSpec``.

        Args:
            spec: Specification of the cube in the block graph.
            block_temporal_height: the number of rounds of stabilizer measurements
            (ignoring one layer for initialization and another for final measurement).

        Returns:
            a ``Block`` based on the provided ``CubeSpec``.

        """
        ...


class PipeBuilder(Protocol):
    """Protocol for building a `Block` based on a `PipeSpec`."""

    def __call__(self, spec: PipeSpec, block_temporal_height: LinearFunction) -> Block:
        """Build a `CompiledBlock` instance from a `PipeSpec`.

        Args:
            spec: Specification of the cube in the block graph.
            block_temporal_height: the number of rounds of stabilizer measurements
            (ignoring one layer for initialization and another for final measurement).

        Returns:
            a `CompiledBlock` based on the provided `PipeSpec`.

        """
        ...


@dataclass(frozen=True)
class PipeSpec:
    """Specification of a pipe in a block graph.

    The `PipeSpec` is used to determine the substitution rules between the two
    `CompiledBlock`s connected by the pipe. The substitution rules are used to
    update the layers of the `CompiledBlock`s based on the plaquettes in the
    `Substitution`.

    Attributes:
        cube_specs: the ordered cube specifications. By convention, the cube
            corresponding to ``cube_specs[0]`` should have a smaller position
            than the cube corresponding to ``cube_specs[1]``.
        cube_templates: templates used to implement the respective entry in
            ``cube_specs``.
        pipe_type: the type of the pipe connecting the two cubes.
        has_spatial_up_or_down_pipe_in_timeslice: a flag indicating if a spatial
            pipe at the top or bottom of a spatial cube is executed on the same
            timeslice as this cube. This information is needed for the fixed
            boundary convention.
        at_temporal_hadamard_layer: flag indicating whether the pipe is a temporal
            pipe and there is a temporal Hadamard pipe at the same Z position
            in the block graph.

    """

    cube_specs: tuple[CubeSpec, CubeSpec]
    cube_templates: tuple[RectangularTemplate, RectangularTemplate]
    pipe_kind: PipeKind
    has_spatial_up_or_down_pipe_in_timeslice: bool = False
    at_temporal_hadamard_layer: bool = False



from __future__ import annotations

from collections.abc import Sequence
from dataclasses import dataclass
from typing import Literal

import stim

from tqec.circuit.moment import Moment
from tqec.circuit.qubit import GridQubit
from tqec.circuit.schedule.circuit import ScheduledCircuit
from tqec.circuit.schedule.schedule import Schedule
from tqec.compile.specs.library.generators.constants import EXTENDED_PLAQUETTE_SCHEDULES
from tqec.plaquette.plaquette import Plaquette
from tqec.plaquette.qubit import PlaquetteQubits
from tqec.plaquette.rpng.rpng import RPNG, PauliBasis, RPNGDescription
from tqec.utils.enums import Basis
from tqec.utils.exceptions import TQECError
from tqec.utils.instructions import MEASUREMENT_INSTRUCTION_NAMES, RESET_INSTRUCTION_NAMES


def _get_spatial_cube_arm_name(
    basis_left: PauliBasis | None,
    basis_right: PauliBasis | None,
    position: Literal["UP", "DOWN"],
    reset: Basis | None,
    measurement: Basis | None,
    is_reverse: bool,
) -> str:
    parts = ["SpatialCubeArm", position]
    if basis_left is not None:
        parts.append("L" + basis_left.name.upper())
    if basis_right is not None:
        parts.append("R" + basis_right.name.upper())
    if reset is not None:
        parts.append(f"R{reset.value.upper()}")
    if measurement is not None:
        parts.append(f"M{measurement.value.upper()}")
    if is_reverse:
        parts.append("reversed")
    return "_".join(parts)


def _check_schedules(
    schedules: Sequence[int | None], first_available_schedule: int, last_available_schedule: int
) -> None:
    considered_schedules: list[int] = [s for s in schedules if s is not None]
    for s in considered_schedules:
        if not (first_available_schedule <= s <= last_available_schedule):
            raise TQECError(
                f"Got a schedule of {s} for a part of an extended plaquette. With the current "
                f"implementation, that schedule should be between {first_available_schedule} and "
                f"{last_available_schedule} (both inclusive)."
            )
    if len(frozenset(considered_schedules)) != len(considered_schedules):
        raise TQECError(
            "Found a duplicated schedule in the provided schedules. That would lead to a qubit "
            "being used twice in the same timestep, which is not supported."
        )


def _make_spatial_cube_arm_memory_plaquette_up(
    left_qubit: RPNG,
    right_qubit: RPNG,
    reset: Basis | None = None,
    measurement: Basis | None = None,
    is_reversed: bool = False,
) -> Plaquette:
    # Checking the validity of the provided schedules
    first_available_schedule = 2 if not is_reversed else 3
    last_available_schedule = 4 if not is_reversed else 5
    _check_schedules(
        [left_qubit.n, right_qubit.n], first_available_schedule, last_available_schedule
    )
    # dl ---- dr
    # |        |
    # |   s1   |
    # |        |
    # s2 ---- s2
    qubits = PlaquetteQubits(
        [GridQubit(-1, -1), GridQubit(1, -1)],
        [GridQubit(0, 0), GridQubit(1 if is_reversed else -1, 1)],
    )
    dl, dr = tuple(qubits.data_qubits_indices)
    s1, s2 = tuple(qubits.syndrome_qubits_indices)
    # Define the base moments, only containing the reset on s2 as that is the
    # only operation that does not depend on the parameters of this function.
    base_moments = [
        Moment(stim.Circuit(f"RZ {s2}")),
        Moment(stim.Circuit()),
        Moment(stim.Circuit()),
        Moment(stim.Circuit()),
        Moment(stim.Circuit()),
        Moment(stim.Circuit()),
        Moment(stim.Circuit()),
        Moment(stim.Circuit()),
    ]
    # Add the GHZ state creation and measurement.
    if not is_reversed:
        base_moments[0].append("RX", [s1], [])
        base_moments[1].append("CX", [s1, s2], [])
        base_moments[5].append("CX", [s2, s1], [])
    else:
        base_moments[1].append("RZ", [s1], [])
        base_moments[2].append("CX", [s2, s1], [])
        base_moments[6].append("CX", [s1, s2], [])
        base_moments[7].append("MX", [s1], [])
    # Add controlled gates
    if left_qubit.p is not None and left_qubit.n is not None:
        base_moments[left_qubit.n].append(f"C{left_qubit.p.name.upper()}", [s1, dl], [])
    if right_qubit.p is not None and right_qubit.n is not None:
        base_moments[right_qubit.n].append(f"C{right_qubit.p.name.upper()}", [s1, dr], [])
    # Add data-qubit reset/measurement if needed
    # Note about resets: data-qubits (i.e., the 4 corners) are already in a
    # correct state and we should not reset them. Internal qubits are also
    # already reset individually by the circuit constructed above. That means
    # that we should NOT reset anything here. Nevertheless, the reset argument
    # is kept because the plaquette naming should be adapted.
    if measurement:
        # Add ancilla measurements if data-qubits are measured.
        base_moments[-1].append("M", [s2] if is_reversed else [s1, s2], [])
    # Finally, return the plaquette
    return Plaquette(
        _get_spatial_cube_arm_name(
            left_qubit.p, right_qubit.p, "UP", reset, measurement, is_reversed
        ),
        qubits,
        ScheduledCircuit(base_moments, 0, qubits.qubit_map),
        MEASUREMENT_INSTRUCTION_NAMES | RESET_INSTRUCTION_NAMES,
    )


def _make_spatial_cube_arm_memory_plaquette_down(
    left_qubit: RPNG,
    right_qubit: RPNG,
    reset: Basis | None = None,
    measurement: Basis | None = None,
    is_reversed: bool = False,
) -> Plaquette:
    # Checking the validity of the provided schedules
    first_available_schedule = 3 if not is_reversed else 2
    last_available_schedule = 5 if not is_reversed else 4
    _check_schedules(
        [left_qubit.n, right_qubit.n], first_available_schedule, last_available_schedule
    )
    # s2 ---- s2
    # |        |
    # |   s1   |
    # |        |
    # dl ---- dr
    qubits = PlaquetteQubits(
        [GridQubit(-1, 1), GridQubit(1, 1)],
        [GridQubit(0, 0), GridQubit(1 if is_reversed else -1, -1)],
    )
    dl, dr = tuple(qubits.data_qubits_indices)
    s1, s2 = tuple(qubits.syndrome_qubits_indices)
    # Define the base moments, only containing the reset on s2 as that is the
    # only operation that does not depend on the parameters of this function.
    base_moments = [
        Moment(stim.Circuit(f"RZ {s2}")),
        Moment(stim.Circuit()),
        Moment(stim.Circuit()),
        Moment(stim.Circuit()),
        Moment(stim.Circuit()),
        Moment(stim.Circuit()),
        Moment(stim.Circuit()),
        Moment(stim.Circuit()),
    ]
    # Add the GHZ state creation and measurement.
    if is_reversed:
        base_moments[0].append("RX", [s1], [])
        base_moments[1].append("CX", [s1, s2], [])
        base_moments[5].append("CX", [s2, s1], [])
    else:
        base_moments[1].append("RZ", [s1], [])
        base_moments[2].append("CX", [s2, s1], [])
        base_moments[6].append("CX", [s1, s2], [])
        base_moments[7].append("MX", [s1], [])
    # Add controlled gates
    if left_qubit.p is not None and left_qubit.n is not None:
        base_moments[left_qubit.n].append(f"C{left_qubit.p.name.upper()}", [s1, dl], [])
    if right_qubit.p is not None and right_qubit.n is not None:
        base_moments[right_qubit.n].append(f"C{right_qubit.p.name.upper()}", [s1, dr], [])
    # Add data-qubit reset/measurement if needed
    # Note about resets: data-qubits (i.e., the 4 corners) are already in a
    # correct state and we should not reset them. Internal qubits are also
    # already reset individually by the circuit constructed above. That means
    # that we should NOT reset anything here. Nevertheless, the reset argument
    # is kept because the plaquette naming should be adapted.
    if measurement:
        # Add ancilla measurements if data-qubits are measured.
        base_moments[-1].append("M", [s1, s2] if is_reversed else [s2], [])
    # Finally, return the plaquette
    return Plaquette(
        _get_spatial_cube_arm_name(
            left_qubit.p, right_qubit.p, "DOWN", reset, measurement, is_reversed
        ),
        qubits,
        ScheduledCircuit(base_moments, 0, qubits.qubit_map),
        MEASUREMENT_INSTRUCTION_NAMES | RESET_INSTRUCTION_NAMES,
    )


def get_extended_plaquette(
    rpng: RPNGDescription,
    reset: Basis | None = None,
    measurement: Basis | None = None,
    is_reversed: bool = False,
) -> tuple[Plaquette, Plaquette]:
    """Create an extended plaquette from the provided RPNG description.

    Args:
        rpng: description of the 4 corners of the extended plaquette.
        reset: basis of the reset operation performed on internal data-qubits (used as syndrome
            qubits). Defaults to ``None`` that translates to no reset being applied on data-qubits.
        measurement: basis of the measurement operation performed on internal data-qubits (used as
            syndrome qubits). Defaults to ``None`` that translates to no measurement being applied
            on data-qubits.
        is_reversed: flag indicating if the plaquette schedule should be reversed or not. Useful to
            limit the loss of code distance when hook errors are not correctly oriented by
            alternating regular and reversed plaquettes.

    Returns:
        a pair of plaquettes ``(UP, DOWN)`` implementing the extended stabilizer.

    """
    tl, tr, bl, br = rpng.corners
    return (
        _make_spatial_cube_arm_memory_plaquette_up(tl, tr, reset, measurement, is_reversed),
        _make_spatial_cube_arm_memory_plaquette_down(bl, br, reset, measurement, is_reversed),
    )


@dataclass(frozen=True)
class ExtendedPlaquette:
    top: Plaquette
    bottom: Plaquette


@dataclass(frozen=True)
class ExtendedPlaquetteCollection:
    bulk: ExtendedPlaquette
    left_with_arm: ExtendedPlaquette
    left_without_arm: ExtendedPlaquette
    right_with_arm: ExtendedPlaquette
    right_without_arm: ExtendedPlaquette

    @staticmethod
    def from_description(
        description: RPNGDescription,
        reset: Basis | None,
        measurement: Basis | None,
        is_reversed: bool,
    ) -> ExtendedPlaquetteCollection:
        """Build an instance from the provided ``RPNGDescription``."""
        up, down = get_extended_plaquette(description, reset, measurement, is_reversed)
        # In the calls to project_on_data_qubit_indices, it is important to remember
        # that individual plaquettes composing the extended plaquette have slightly
        # unconventional qubit layouts. In the below ASCII representation, "s" means
        # "syndrome qubit", "d" means "data qubit", the numbers are the respective
        # qubit indices. "s3" appears at two places because it swaps between the
        # two locations depending on whether the plaquette is reversed or not.
        # UP:
        # d0 ---- d1
        # |        |
        # |   s2   |
        # |        |
        # s3 ---- s3
        # DOWN:
        # s3 ---- s3
        # |        |
        # |   s2   |
        # |        |
        # d0 ---- d1
        return ExtendedPlaquetteCollection(
            bulk=ExtendedPlaquette(up, down),
            left_with_arm=ExtendedPlaquette(up.project_on_data_qubit_indices([1]), down),
            left_without_arm=ExtendedPlaquette(
                up.project_on_data_qubit_indices([1]), down.project_on_data_qubit_indices([1])
            ),
            right_with_arm=ExtendedPlaquette(up, down.project_on_data_qubit_indices([0])),
            right_without_arm=ExtendedPlaquette(
                up.project_on_data_qubit_indices([0]), down.project_on_data_qubit_indices([0])
            ),
        )

    @staticmethod
    def from_basis(
        basis: Basis,
        reset: Basis | None,
        measurement: Basis | None,
        is_reversed: bool,
        schedule: Sequence[int] | Schedule | None = None,
    ) -> ExtendedPlaquetteCollection:
        """Create an instance from a basis and a schedule.

        This allows one to create an extended plaquette that measures its four corners in the
        provided basis, with the provided schedule. It does not allow mixed-basis plaquette, and as
        such it is a strictly less capable version of :meth:`from_description`, but it might be more
        convenient to use.

        Args:
            basis: stabilizer that will be measured on all the corners of the returned extended
                stabilizers.
            reset: basis of the reset operation performed on internal data-qubits (used as syndrome
                qubits). Defaults to ``None`` that translates to no reset being applied on
                data-qubits.
            measurement: basis of the measurement operation performed on internal data-qubits (used
                as syndrome qubits). Defaults to ``None`` that translates to no measurement being
                applied on data-qubits.
            is_reversed: flag indicating if the plaquette schedule should be reversed or not. Useful
                to limit the loss of code distance when hook errors are not correctly oriented by
                alternating regular and reversed plaquettes.
            schedule: if provided, the returned extended plaquettes will unconditionally implement
                this schedule (no matter the provided value of ``is_reversed``, it is up to the
                caller to ensure the provided value is valid). If not provided, a default schedule
                that depends on ``is_reversed`` is used. Needs to contain exactly 4 integer entries.

        Returns:
            a collection of extended plaquettes measuring the provided basis on data-qubits in the
            given schedule.

        """
        if schedule is None:
            schedule = EXTENDED_PLAQUETTE_SCHEDULES[is_reversed]
        # Not including reset and measurement in the RPNG description as these are specially placed
        # for an extended plaquette
        description = RPNGDescription.from_basis_and_schedule(basis, schedule)
        return ExtendedPlaquetteCollection.from_description(
            description, reset, measurement, is_reversed
        )



from itertools import chain, repeat
from typing import TypeGuard, cast

import numpy

from tqec.compile.blocks.layers.atomic.base import BaseLayer
from tqec.compile.blocks.layers.atomic.layout import LayoutLayer
from tqec.compile.blocks.layers.composed.base import BaseComposedLayer
from tqec.compile.blocks.layers.composed.repeated import RepeatedLayer
from tqec.compile.blocks.layers.composed.sequenced import SequencedLayers
from tqec.compile.blocks.positioning import LayoutPosition2D
from tqec.utils.exceptions import TQECError
from tqec.utils.scale import PhysicalQubitScalable2D, round_or_fail

# Note on the few functions below:
# These functions are needed mostly for typing purposes. The main issue is that
# all the type-checkers I tested are not able to infer the correct type in the
# situation below:
#
# obj: dict[int, int | float] = {}
# if all(isinstance(value, float) for value in obj.values()):
#     # We expect obj to be "dict[int, float]", but type checkers are not able
#     # to infer that.
#
# To circumvent that, the functions below use typing.TypeGuard (see
# https://docs.python.org/3/library/typing.html#typing.TypeGuard).


def contains_only_base_layers(
    layers: dict[LayoutPosition2D, BaseLayer | BaseComposedLayer],
) -> TypeGuard[dict[LayoutPosition2D, BaseLayer]]:
    """Ensure correct typing when used in a conditional block."""
    return all(isinstance(layer, BaseLayer) for layer in layers.values())


def contains_only_composed_layers(
    layers: dict[LayoutPosition2D, BaseLayer | BaseComposedLayer],
) -> TypeGuard[dict[LayoutPosition2D, BaseComposedLayer]]:
    """Ensure correct typing when used in a conditional block."""
    return all(isinstance(layer, BaseComposedLayer) for layer in layers.values())


def contains_only_repeated_layers(
    layers: dict[LayoutPosition2D, BaseComposedLayer],
) -> TypeGuard[dict[LayoutPosition2D, RepeatedLayer]]:
    """Ensure correct typing when used in a conditional block."""
    return all(isinstance(layer, RepeatedLayer) for layer in layers.values())


def contains_only_sequenced_layers(
    layers: dict[LayoutPosition2D, BaseComposedLayer],
) -> TypeGuard[dict[LayoutPosition2D, SequencedLayers]]:
    """Ensure correct typing when used in a conditional block."""
    return all(isinstance(layer, SequencedLayers) for layer in layers.values())


def contains_only_repeated_or_sequenced_layers(
    layers: dict[LayoutPosition2D, BaseComposedLayer],
) -> TypeGuard[dict[LayoutPosition2D, SequencedLayers | RepeatedLayer]]:
    """Ensure correct typing when used in a conditional block."""
    return all(isinstance(layer, (SequencedLayers, RepeatedLayer)) for layer in layers.values())


def merge_base_layers(
    layers: dict[LayoutPosition2D, BaseLayer],
    scalable_qubit_shape: PhysicalQubitScalable2D,
) -> LayoutLayer:
    """Merge several :class:`.BaseLayer` instances into one."""
    return LayoutLayer(layers, scalable_qubit_shape)


def merge_composed_layers(
    layers: dict[LayoutPosition2D, BaseComposedLayer],
    scalable_qubit_shape: PhysicalQubitScalable2D,
) -> BaseComposedLayer:
    """Merge several :class:`.BaseComposedLayer` instances into one.

    The specific type returned will depend on the provided ``layers``.

    """
    # First, check that all the provided layers have the same scalable timesteps.
    different_timesteps = frozenset(layer.scalable_timesteps for layer in layers.values())
    if len(different_timesteps) > 1:
        raise TQECError(
            "Cannot merged BaseComposedLayer instances that have different lengths. "
            f"Found the following different lengths: {different_timesteps}."
        )
    if contains_only_repeated_layers(layers):
        repeated_layers = cast(dict[LayoutPosition2D, RepeatedLayer], layers)
        return merge_repeated_layers(repeated_layers, scalable_qubit_shape)
    if contains_only_sequenced_layers(layers):
        sequenced_layers = cast(dict[LayoutPosition2D, SequencedLayers], layers)
        return merge_sequenced_layers(sequenced_layers, scalable_qubit_shape)
    # We are left here with a mix of RepeatedLayer and SequencedLayers.
    # Check that, in case a new subclass of BaseComposedLayer has been introduced.
    if not contains_only_repeated_or_sequenced_layers(layers):
        unknown_types = {type(layer) for layer in layers.values()} - {
            RepeatedLayer,
            SequencedLayers,
        }
        unknown_types_names = {t.__name__ for t in unknown_types}
        raise NotImplementedError(
            f"Found instances of {unknown_types_names} that are not yet "
            "implemented in _merge_composed_layers."
        )
    mixed_layers = cast(dict[LayoutPosition2D, SequencedLayers | RepeatedLayer], layers)
    return merge_repeated_and_sequenced_layers(mixed_layers, scalable_qubit_shape)


def merge_repeated_layers(
    layers: dict[LayoutPosition2D, RepeatedLayer],
    scalable_qubit_shape: PhysicalQubitScalable2D,
) -> RepeatedLayer:
    """Merge several RepeatedLayer that should be executed in parallel.

    Args:
        layers: the different repeated layers that should be merged.
        scalable_qubit_shape: scalable shape of a scalable qubit. Considered
            valid across the whole domain.

    Raises:
        TQECError: if the provided repeated layers do not all have the same
            temporal footprint.
        NotImplementedError: if any of the provided repeated layers have an
            internal layer (i.e., the layer that is being repeated) with a
            non-constant temporal footprint.

    Returns:
        a unique repeated layer implementing the same piece of computation as
        the provided repeated layers.

    """
    # First, check that all the provided layers have the same scalable timesteps.
    different_timesteps = frozenset(layer.scalable_timesteps for layer in layers.values())
    if len(different_timesteps) > 1:
        raise TQECError(
            "Cannot merge RepeatedLayer instances that have different lengths. "
            f"Found the following different lengths: {different_timesteps}."
        )
    scalable_timesteps = next(iter(different_timesteps))
    timesteps_per_repetition: dict[LayoutPosition2D, int] = {}
    for pos, layer in layers.items():
        timesteps = layer.internal_layer.scalable_timesteps
        # Implementation note: the fact that the internal layer is of constant
        # size in the time dimension is used later in the function.
        if not timesteps.is_constant():
            raise NotImplementedError(
                "Found a RepeatedLayer with a scalable in time internal layer. "
                "This is currently not supported."
            )
        timesteps_per_repetition[pos] = round_or_fail(timesteps.offset)

    # The internal layer of the returned RepeatedLayer should be divisible by
    # all the values in timesteps_per_repetition. Basically we want the smallest
    # integer that is a multiple of each timesteps, i.e., the "least common
    # multiple".
    # Also, `1`s do not have to be accounted for because they are trivially a
    # divisor of any integer.
    considered_timesteps = [
        timesteps for timesteps in timesteps_per_repetition.values() if timesteps > 1
    ]
    # If we only have `1`s (and so considered_timesteps is empty), that's trivial:
    if not considered_timesteps:
        # Sanity check on repetitions
        different_repetitions = frozenset(layer.scalable_timesteps for layer in layers.values())
        assert len(different_repetitions) == 1
        # Sanity check on types: SequencedLayer guarantees that it contains at
        # least 2 base layers, so we cannot have any SequencedLayer instance here,
        # meaning that we only have PlaquetteLayer instances.
        inner_layers = {pos: layer.internal_layer for pos, layer in layers.items()}
        assert contains_only_base_layers(inner_layers)
        return RepeatedLayer(
            merge_base_layers(
                cast(dict[LayoutPosition2D, BaseLayer], inner_layers), scalable_qubit_shape
            ),
            next(iter(different_repetitions)),
        )
    # Else, we need the least common multiple
    num_internal_layers = numpy.lcm.reduce(considered_timesteps)
    # And we create sequences of that size to merge them!
    base_sequences: dict[LayoutPosition2D, list[BaseLayer]] = {}
    for pos, layer in layers.items():
        internal_layer = layer.internal_layer
        if isinstance(internal_layer, BaseLayer):
            base_sequences[pos] = [internal_layer for _ in range(num_internal_layers)]
        else:
            # We know for sure that the internal layer is of constant size so
            # we can get its layers for any value of k we want.
            layer_sequence = list(internal_layer.all_layers(0))
            base_len = len(layer_sequence)
            internal_repetitions = num_internal_layers // base_len
            base_sequences[pos] = list(
                chain.from_iterable(repeat(layer_sequence, internal_repetitions))
            )
    # Checking post-condition of the above loop.
    assert all(
        len(layer_sequence) == num_internal_layers for layer_sequence in base_sequences.values()
    )
    # Computing the new scalable repetitions number.
    # Note that the following should in theory never fail, because all the
    # repeated layers given in input have the SAME scalable shapes, and so in
    # theory num_internal_layers should be a divisor of each of the overall
    # scalable shapes.
    repetitions = scalable_timesteps.exact_integer_div(num_internal_layers)
    return RepeatedLayer(
        SequencedLayers(
            [
                merge_base_layers(
                    {pos: layer_sequence[i] for pos, layer_sequence in base_sequences.items()},
                    scalable_qubit_shape,
                )
                for i in range(num_internal_layers)
            ]
        ),
        repetitions,
    )


def merge_sequenced_layers(
    layers: dict[LayoutPosition2D, SequencedLayers],
    scalable_qubit_shape: PhysicalQubitScalable2D,
) -> SequencedLayers:
    """Merge several SequencedLayers that should be executed in parallel.

    Args:
        layers: the different sequenced layers that should be merged.
        scalable_qubit_shape: scalable shape of a scalable qubit. Considered
            valid across the whole domain.

    Raises:
        NotImplementedError: if any of the provided sequenced layer contains a different sub-layer
            schedule.

    Returns:
        a unique sequenced layer implementing the same piece of computation as
        the provided sequenced layers.

    """
    internal_layers_schedules = frozenset(
        sequenced_layer.schedule for sequenced_layer in layers.values()
    )
    if len(internal_layers_schedules) != 1:
        raise NotImplementedError(
            "_merge_sequenced_layers only supports merging sequences that have "
            "layers with a matching temporal schedule. Found the following "
            "different temporal schedules in the provided sequences: "
            f"{internal_layers_schedules}."
        )
    internal_layers_schedule = next(iter(internal_layers_schedules))
    merged_layers: list[LayoutLayer | BaseComposedLayer] = []
    for i in range(len(internal_layers_schedule)):
        layers_at_timestep = {
            pos: sequenced_layers.layer_sequence[i] for pos, sequenced_layers in layers.items()
        }
        if contains_only_base_layers(layers_at_timestep):
            merged_layers.append(
                merge_base_layers(
                    cast(dict[LayoutPosition2D, BaseLayer], layers_at_timestep),
                    scalable_qubit_shape,
                )
            )
        elif contains_only_composed_layers(layers_at_timestep):
            merged_layers.append(
                merge_composed_layers(
                    cast(dict[LayoutPosition2D, BaseComposedLayer], layers_at_timestep),
                    scalable_qubit_shape,
                )
            )
        else:
            raise RuntimeError(
                f"Found a mix of {BaseLayer.__name__} instances and "
                f"{BaseComposedLayer.__name__} instances in a single temporal "
                f"layer. This should be already checked before. This is a "
                "logical error in the code, please open an issue. Found layers:"
                "\n - " + "\n - ".join(repr(layer) for layer in layers_at_timestep.values())
            )
    return SequencedLayers(merged_layers)


def merge_repeated_and_sequenced_layers(
    layers: dict[LayoutPosition2D, SequencedLayers | RepeatedLayer],
    scalable_qubit_shape: PhysicalQubitScalable2D,
) -> SequencedLayers:
    """Merge composed layers with both RepeatedLayer and SequencedLayers instances.

    Raises:
        TQECError: if there is no layer of type SequencedLayers.
        TQECError: if there is no layer of type RepeatedLayer.
        TQECError: if the provided layers have different durations.
        NotImplementedError: if the ScheduledLayers instances in ``layers`` have
            different schedules.

    """
    layer_types = frozenset(type(layer) for layer in layers.values())
    if layer_types != frozenset((RepeatedLayer, SequencedLayers)):
        raise TQECError(
            "Wrong layer types: expecting at least one layer for each of the "
            f"expected types ({RepeatedLayer.__name__} and {SequencedLayers.__name__}) "
            "but got the following types: " + ",".join(t.__name__ for t in layer_types)
        )
    different_timesteps = frozenset(layer.scalable_timesteps for layer in layers.values())
    if len(different_timesteps) > 1:
        raise TQECError(
            f"Cannot merge {RepeatedLayer.__name__} and {SequencedLayers.__name__} "
            "instances that have different durations. Found the following "
            f"different durations: {different_timesteps}."
        )
    sequenced_schedules = frozenset(
        layer.schedule for layer in layers.values() if isinstance(layer, SequencedLayers)
    )
    if len(sequenced_schedules) != 1:
        raise NotImplementedError(
            f"Merging different {SequencedLayers.__name__} instances with different "
            "schedules is not implemented yet. Found the following schedules: "
            f"{sequenced_schedules}."
        )
    schedule = next(iter(sequenced_schedules))
    return merge_sequenced_layers(
        {pos: layer.to_sequenced_layer_with_schedule(schedule) for pos, layer in layers.items()},
        scalable_qubit_shape,
    )



"""Defines some helper functions to plot into insets."""

# required for "3d" projection even though not explicitly used
import sinter
from matplotlib.axes import Axes
from matplotlib.ticker import ScalarFormatter
from mpl_toolkits.mplot3d.axes3d import Axes3D

from tqec.computation.correlation import CorrelationSurface
from tqec.interop.pyzx.plot import (
    draw_correlation_surface_on,
    draw_positioned_zx_graph_on,
)
from tqec.interop.pyzx.positioned import PositionedZX


def add_inset_axes3d(ax_target: Axes, bounds: tuple[float, float, float, float]) -> Axes3D:
    """Wrap ``fig.add_axes`` to achieve ``ax.inset_axes`` functionality for 3D."""
    ax = ax_target.inset_axes(bounds, projection="3d")
    assert isinstance(ax, Axes3D)
    return ax


def plot_observable_as_inset(
    ax_target: Axes,
    zx_graph: PositionedZX,
    observable: CorrelationSurface,
    bounds: tuple[float, float, float, float] = (0.55, 0.0, 0.45, 0.45),
) -> None:
    """Plot the provided observable as an inset in the provided ax.

    Args:
        ax_target: the ax to insert an inset in to plot the correlation surface.
        bounds: (x0, y0, width, height) where (x0, y0) is the position of the
            lower-left corner of the inset. All coordinates are normalized to
            [0, 1] meaning that an input of (0, 0, 1, 1) will span the whole
            `ax_target`.
        zx_graph: ZX graph used.
        observable: correlation surface over the provided `zx_graph` to
            draw.

    """
    inset_ax = add_inset_axes3d(ax_target, bounds)
    inset_ax.set_axis_off()
    draw_positioned_zx_graph_on(zx_graph, inset_ax, node_size=50)
    draw_correlation_surface_on(observable, zx_graph, inset_ax)
    inset_ax.set_facecolor((0.0, 0.0, 0.0, 0.0))


def plot_threshold_as_inset(
    ax_target: Axes,
    stats: list[sinter.TaskStats],
    zoom_bounds: tuple[float, float, float, float],
    threshold: float | None = None,
    inset_bounds: tuple[float, float, float, float] = (0.53, 0.45, 0.4, 0.4),
) -> None:
    """Add an inset plot to ``ax_target`` with the provided ``stats``.

    Args:
        ax_target: ax on which to insert the inset.
        stats: ``sinter`` statistics to plot with ``sinter.plot_error_rate`` on
            the created inset.
        zoom_bounds: ``(x0, y0, x1, y1)``. The inset will have physical
            error-rate (X-axis coordinate) ranging from ``x0`` to ``x1`` and
            logical error-rate (Y-axis coordinate) ranging from ``y0` to ``y1``.
            Note that you likely want ``y1 > y0`` to have an inset plot that
            looks like the main error-rate plot.
        threshold: if not ``None``, a red horizontal line is added to the inset
            plot to the physical error-rate given in this argument.
        inset_bounds: bounds in axes coordinates (i.e., between 0 and 1) for the
            rectangle in which the inset should appear on the main axes. The
            default value tries to place the inset such that it does not overlap
            with the logical error-rate curves on the main plot while still
            leaving room to plot the observable in another inset.

    """
    # Creating the inset
    inset_ax = ax_target.inset_axes(
        inset_bounds,
        xlim=(zoom_bounds[0], zoom_bounds[2]),
        ylim=(zoom_bounds[1], zoom_bounds[3]),
    )
    # Configuring the inset to look nice
    # 1. This is a loglog plot.
    inset_ax.loglog()
    # 2. Ticks management.
    # 2.1. Because we expect the logical error-rate at the threshold to be high
    #      we can format the Y-axis ticks using that knowledge.
    inset_ax.invert_yaxis()
    inset_ax.yaxis.set_label_position("right")
    inset_ax.yaxis.tick_right()
    inset_ax.yaxis.set_minor_formatter(ScalarFormatter())
    inset_ax.yaxis.set_major_formatter(ScalarFormatter())
    inset_ax.tick_params(axis="y", which="minor", labelsize=7)
    inset_ax.tick_params(axis="y", which="major", labelsize=7)
    # 2.2. The physical error-rate at which the threshold is located should be
    #      quite small (below 10**-2 from experiments), so keep the scientific
    #      notation here.
    inset_ax.tick_params(axis="x", which="major", labelsize=7)
    inset_ax.tick_params(axis="x", which="minor", labelsize=7)
    # 3. Make the grid apparent.
    inset_ax.grid(which="both", axis="both")
    # 4. Annotate the inset correctly.
    _, (lower_left, upper_left, lower_right, upper_right) = ax_target.indicate_inset_zoom(
        inset_ax, edgecolor="black", alpha=0.8
    )  # type: ignore
    lower_left.set_visible(True)  # type: ignore
    upper_right.set_visible(True)  # type: ignore
    lower_right.set_visible(False)  # type: ignore
    upper_left.set_visible(False)  # type: ignore

    # Plotting the data
    sinter.plot_error_rate(
        ax=inset_ax,
        stats=stats,
        x_func=lambda stat: stat.json_metadata["p"],
        group_func=lambda stat: stat.json_metadata["d"],
        plot_args_func=lambda index, group_key, group_stats: {"markersize": 3},
    )
    # Drawing the threshold if provided
    if threshold is not None:
        inset_ax.axvline(threshold, color="r")
        inset_ax.text(
            threshold * 1.02,
            zoom_bounds[1] * 0.96,
            f"{threshold:.3g}",
            fontsize=6,
            color="r",
            horizontalalignment="left",
            verticalalignment="top",
        )



import math

import svg

from tqec.circuit.qubit import GridQubit
from tqec.compile.observables.builder import Observable
from tqec.visualisation.configuration import DrawerConfiguration


def _get_observable_star_svg(
    r_outer: float,
    r_inner: float,
    fill: str = "none",
    fill_opacity: float = 0.5,
    stroke_color: str = "red",
    stroke_width: float = 0.5,
) -> svg.Polygon:
    points: list[tuple[float, float]] = []
    n = 5
    angle = math.pi / n
    for i in range(2 * n):
        r = r_outer if i % 2 == 0 else r_inner
        theta = i * angle - math.pi / 2
        x = r * math.cos(theta)
        y = r * math.sin(theta)
        points.append((x, y))
    return svg.Polygon(
        points=points,  # type: ignore[arg-type]
        fill=fill,
        stroke=stroke_color,
        stroke_width=stroke_width,
        fill_opacity=fill_opacity,
    )


def get_observable_svg(
    observable: Observable,
    top_left_qubit: GridQubit,
    plaquette_width: float,
    plaquette_height: float,
    configuration: DrawerConfiguration = DrawerConfiguration(),
) -> svg.G:
    """Return an SVG element with the provided ``observable`` drawn and a transparent background.

    Args:
        observable: the observable to plot. It is represented as a list of qubits
            whose measurements are included in the observable.
        top_left_qubit: coordinates of the qubit at the very top-left of the
            visualisation canvas. Used to correctly offset qubit values from the
            provided ``errors``.
        plaquette_width: width (in SVG dimensions) of a regular square plaquette.
        plaquette_height: height (in SVG dimensions) of a regular square plaquette.
        configuration: drawing configuration.

    Returns:
        a SVG element containing as many sub-elements as there are measurements in the
        provided ``observable``.

    """
    width_between_qubits: float = plaquette_width / 2
    height_between_qubits: float = plaquette_height / 2

    layer: list[svg.Element] = []

    for q in observable.measured_qubits:
        qx, qy = q.x, q.y
        color = configuration.observable_star_color.rgba.to_hex()
        # Make the coordinates relative to the top-left qubit.
        qx -= top_left_qubit.x
        qy -= top_left_qubit.y
        # Plot the star for a single-qubit measurement.
        star_svg = _get_observable_star_svg(
            r_outer=0.15 * plaquette_width,
            r_inner=0.07 * plaquette_width,
            fill=color,
            fill_opacity=configuration.observable_fill_opacity,
            stroke_color=color,
            stroke_width=configuration.observable_stroke_width_multiplier * plaquette_width,
        )
        layer.append(
            svg.G(
                elements=[star_svg],
                transform=[svg.Translate(qx * width_between_qubits, qy * height_between_qubits)],
            )
        )
    return svg.G(elements=layer)



from __future__ import annotations

from collections.abc import Iterable, Mapping
from functools import cached_property
from typing import Final, cast

from typing_extensions import override

from tqec.compile.blocks.enums import SpatialBlockBorder, TemporalBlockBorder
from tqec.compile.blocks.layers.atomic.base import BaseLayer
from tqec.compile.blocks.layers.atomic.layout import LayoutLayer
from tqec.compile.blocks.layers.composed.base import BaseComposedLayer
from tqec.compile.blocks.layers.composed.sequenced import SequencedLayers
from tqec.compile.blocks.layers.merge import (
    contains_only_base_layers,
    contains_only_composed_layers,
    merge_base_layers,
    merge_composed_layers,
)
from tqec.compile.blocks.positioning import LayoutPosition2D
from tqec.utils.exceptions import TQECError
from tqec.utils.scale import LinearFunction, PhysicalQubitScalable2D


class Block(SequencedLayers):
    """Encodes the implementation of a block.

    This data structure is voluntarily very generic. It represents blocks as a
    sequence of layers that can be instances of either
    :class:`~tqec.compile.blocks.layers.atomic.base.BaseLayer` or
    :class:`~tqec.compile.blocks.layers.composed.base.BaseComposedLayer`.

    Depending on the stored layers, this class can be used to represent regular
    cubes (i.e. scaling in the 3 dimensions with ``k``) as well as pipes (i.e.
    scaling in only 2 dimension with ``k``).

    """

    @override
    def with_spatial_borders_trimmed(self, borders: Iterable[SpatialBlockBorder]) -> Block:
        return Block(
            self._layers_with_spatial_borders_trimmed(borders),
            self.trimmed_spatial_borders | frozenset(borders),
        )

    @override
    def with_temporal_borders_replaced(
        self,
        border_replacements: Mapping[TemporalBlockBorder, BaseLayer | None],
    ) -> Block | None:
        if not border_replacements:
            return self
        layers = self._layers_with_temporal_borders_replaced(border_replacements)
        return Block(layers) if layers else None

    def get_atomic_temporal_border(self, border: TemporalBlockBorder) -> BaseLayer:
        """Get the layer at the provided temporal ``border``.

        This method is different to :meth:`get_temporal_layer_on_border` in that it raises when the
        border is not an atomic layer.

        Raises:
            TQECError: if the layer at the provided temporal ``border`` is not atomic (i.e., an
                instance of :class:`.BaseLayer`).

        """
        layer_index: int
        match border:
            case TemporalBlockBorder.Z_NEGATIVE:
                layer_index = 0
            case TemporalBlockBorder.Z_POSITIVE:
                layer_index = -1
        layer = self.layer_sequence[layer_index]
        if not isinstance(layer, BaseLayer):
            raise TQECError(
                "Expected to recover a temporal **border** (i.e. an atomic "
                f"layer) but got an instance of {type(layer).__name__} instead."
            )
        return layer

    @cached_property
    def dimensions(self) -> tuple[LinearFunction, LinearFunction, LinearFunction]:
        """Return the dimensions of ``self``.

        Returns:
            a 3-dimensional tuple containing the width for each of the
            ``(x, y, z)`` dimensions.

        """
        spatial_shape = self.scalable_shape
        return spatial_shape.x, spatial_shape.y, self.scalable_timesteps

    @property
    def is_cube(self) -> bool:
        """Return ``True`` if ``self`` represents a cube, else ``False``.

        A cube is defined as a block with all its 3 dimensions that are scalable.

        """
        return all(dim.is_scalable() for dim in self.dimensions)

    @property
    def is_pipe(self) -> bool:
        """Return ``True`` if ``self`` represents a pipe, else ``False``.

        A pipe is defined as a block with all but one of its 3 dimensions that are scalable.

        """
        return sum(dim.is_scalable() for dim in self.dimensions) == 2

    @property
    def is_temporal_pipe(self) -> bool:
        """Return ``True`` if ``self`` is a temporal pipe, else ``False``.

        A temporal pipe is a pipe (exactly 2 scalable dimensions) for which the non-scalable
        dimension is the third one (time dimension).

        """
        return self.is_pipe and self.dimensions[2].is_constant()

    def __eq__(self, value: object) -> bool:
        return isinstance(value, Block) and super().__eq__(value)

    def __hash__(self) -> int:
        raise NotImplementedError(f"Cannot hash efficiently a {type(self).__name__}.")


def merge_parallel_block_layers(
    blocks_in_parallel: Mapping[LayoutPosition2D, Block],
    scalable_qubit_shape: PhysicalQubitScalable2D,
) -> list[LayoutLayer | BaseComposedLayer]:
    """Merge several stacks of layers executed in parallel into one stack of larger layers.

    Args:
        blocks_in_parallel: a 2-dimensional arrangement of blocks. Each of the
            provided block MUST have the exact same duration (also called
            "temporal footprint", or number of atomic layers).
        scalable_qubit_shape: scalable shape of a scalable qubit. Considered
            valid across the whole domain.

    Returns:
        a stack of layers representing the same slice of computation as the
        provided ``blocks_in_parallel``.

    Raises:
        TQECError: if two items from the provided ``blocks_in_parallel`` do
            not have the same temporal footprint.
        NotImplementedError: if the provided blocks cannot be merged due to a
            code branch not being implemented yet (and not due to a logical
            error making the blocks unmergeable).

    """
    if not blocks_in_parallel:
        return []
    internal_layers_schedules = frozenset(
        tuple(layer.scalable_timesteps for layer in block.layer_sequence)
        for block in blocks_in_parallel.values()
    )
    if len(internal_layers_schedules) != 1:
        raise NotImplementedError(
            "merge_parallel_block_layers only supports merging blocks that have "
            "layers with a matching temporal schedule. Found the following "
            "different temporal schedules in the provided blocks: "
            f"{internal_layers_schedules}."
        )
    schedule: Final = next(iter(internal_layers_schedules))
    merged_layers: list[LayoutLayer | BaseComposedLayer] = []
    for i in range(len(schedule)):
        layers = {pos: block.layer_sequence[i] for pos, block in blocks_in_parallel.items()}
        if contains_only_base_layers(layers):
            merged_layers.append(
                merge_base_layers(
                    cast(dict[LayoutPosition2D, BaseLayer], layers), scalable_qubit_shape
                )
            )
        elif contains_only_composed_layers(layers):
            merged_layers.append(
                merge_composed_layers(
                    cast(dict[LayoutPosition2D, BaseComposedLayer], layers), scalable_qubit_shape
                )
            )
        else:
            raise RuntimeError(
                f"Found a mix of {BaseLayer.__name__} instances and "
                f"{BaseComposedLayer.__name__} instances in a single temporal "
                f"layer. This should be already checked before. This is a "
                "logical error in the code, please open an issue. Found layers:"
                f"\n{list(layers.values())}"
            )
    return merged_layers



"""Block graph that represents a logical stability experiment."""

from tqec.computation.block_graph import BlockGraph
from tqec.utils.enums import Basis
from tqec.utils.position import Position3D


def stability(observable_basis: Basis = Basis.Z) -> BlockGraph:
    """Create a block graph with a single cube that represents a logical stability experiment.

    Args:
        observable_basis: The logical observable basis for the memory experiment.

    Returns:
        A :py:class:`~tqec.computation.block_graph.BlockGraph` instance.

    """
    g = BlockGraph(f"{observable_basis} Basis Stability Experiment")
    node_kind = "ZZX" if observable_basis == Basis.Z else "XXZ"
    g.add_cube(Position3D(0, 0, 0), node_kind)
    return g



from abc import ABC, abstractmethod
from collections.abc import Sequence

import stim
from typing_extensions import override


class InstructionSimplifier(ABC):
    """Base interface for classes to simplify operations.

    Compilation passes often need to introduce new gates in several moments. It
    might happen that these newly inserted gates are already present in the
    moment (which is often the case for the ``H`` gate). This class offers an
    interface to try to simplify those gates when it is possible, before raising
    an exception because the moment is invalid.

    """

    @abstractmethod
    def simplify(
        self, instructions: Sequence[stim.CircuitInstruction]
    ) -> list[stim.CircuitInstruction]:
        """Simplify a list of instructions that are happening at the same moment in the circuit."""
        pass


class NoInstructionSimplification(InstructionSimplifier):
    @override
    def simplify(
        self, instructions: Sequence[stim.CircuitInstruction]
    ) -> list[stim.CircuitInstruction]:
        return list(instructions)


class SelfInverseGateSimplification(InstructionSimplifier):
    def __init__(self, *self_inverse_gates: str):
        """Compilation pass simplifying self-inverse gates when applied more than once."""
        super().__init__()
        self._self_inverse_gates = frozenset(self_inverse_gates)

    @override
    def simplify(
        self, instructions: Sequence[stim.CircuitInstruction]
    ) -> list[stim.CircuitInstruction]:
        # Append in ret all the instructions that are not in
        # self._self_inverse_gates and count the instructions that are in it.
        ret: list[stim.CircuitInstruction] = []
        gate_counter: dict[tuple[str, tuple[stim.GateTarget, ...], tuple[float, ...]], int] = {}
        for instruction in instructions:
            if instruction.name not in self._self_inverse_gates:
                ret.append(instruction)
                continue
            args = tuple(instruction.gate_args_copy())
            for target_group in instruction.target_groups():
                key = (instruction.name, tuple(target_group), args)
                gate_counter[key] = gate_counter.get(key, 0) + 1
        # Instructions in gate_counter are marked as self-inverse. An odd number
        # of such instructions is equivalent to applying the instruction once.
        # An even number is equivalent to not applying the instruction.
        # Not inserting directly into ret yet to group instructions by name in
        # the final instruction list.
        gates_to_apply: dict[tuple[str, tuple[float, ...]], list[stim.GateTarget]] = {}
        for (name, target_group, args), count in gate_counter.items():
            if count % 2 == 1:
                gates_to_apply.setdefault((name, args), []).extend(target_group)
        for (name, args), targets in gates_to_apply.items():
            ret.append(stim.CircuitInstruction(name, targets, args))
        return ret



"""Provides :class:`.ScheduledCircuitTransformationPass`, a generic compilation pass."""

from .schedule import ScheduleConstant as ScheduleConstant
from .schedule import ScheduleFunction as ScheduleFunction
from .schedule import ScheduleOffset as ScheduleOffset
from .transformer import InstructionCreator as InstructionCreator
from .transformer import (
    ScheduledCircuitTransformation as ScheduledCircuitTransformation,
)
from .transformer import (
    ScheduledCircuitTransformationPass as ScheduledCircuitTransformationPass,
)
from .transformer import ScheduledCircuitTransformer as ScheduledCircuitTransformer



from __future__ import annotations

from collections.abc import Iterable
from functools import cached_property
from typing import Final, Literal

from typing_extensions import override

from tqec.compile.blocks.enums import SpatialBlockBorder
from tqec.compile.blocks.layers.atomic.base import BaseLayer
from tqec.compile.blocks.layers.spatial import EXPECTED_SPATIAL_BORDER_WIDTH
from tqec.plaquette.plaquette import Plaquettes
from tqec.templates.base import RectangularTemplate
from tqec.utils.exceptions import TQECError
from tqec.utils.position import Shift2D
from tqec.utils.scale import (
    LinearFunction,
    PhysicalQubitScalable2D,
    PlaquetteScalable2D,
)


class PlaquetteLayer(BaseLayer):
    def __init__(
        self,
        template: RectangularTemplate,
        plaquettes: Plaquettes,
        trimmed_spatial_borders: frozenset[SpatialBlockBorder] = frozenset(),
    ) -> None:
        """Represent a layer with a template and some plaquettes.

        This class implements the layer interface by using a template and some
        plaquettes. This is the preferred way of representing a layer.

        Args:
            template: template used in conjunction to ``plaquettes`` to
                represent the quantum circuit implementing the layer.
            plaquettes: plaquettes used in conjunction to ``template`` to
                represent the quantum circuit implementing the layer.
            trimmed_spatial_borders: all the spatial borders that have been
                removed from the layer.

        Raises:
            TQECError: if the provided ``template`` increments do not match
                with the expected spatial border width.
            TQECError: if the provided ``template`` and
                ``trimmed_spatial_borders`` can lead to empty instantiations for
                ``k >= 1``.

        """
        super().__init__(trimmed_spatial_borders)
        self._template = template
        self._plaquettes = plaquettes
        self._post_init_check()

    def _post_init_check(self) -> None:
        # Shortening variable name for convenience
        _ew: Final[int] = EXPECTED_SPATIAL_BORDER_WIDTH
        expected_shifts = Shift2D(_ew, _ew)
        if (shifts := self._template.get_increments()) != expected_shifts:
            raise TQECError(
                f"Spatial borders are expected to be {_ew} qubits large. Got a "
                f"Template instance with {shifts:=}. Removing a border from "
                f"such a template instance would remove more than {_ew} qubits, "
                "which is not supported."
            )
        # We require the template shape to be strictly positive for any value of
        # k > 0.
        shape = PlaquetteLayer._get_template_shape(self._template, self.trimmed_spatial_borders)
        shape1 = shape.to_numpy_shape(1)
        # Check that the shape is valid (i.e., strictly positive) for k == 1.
        if not all(coord > 0 for coord in shape1):
            raise TQECError(
                "The provided template/trimmed_spatial_borders combo leads to an "
                f"invalid template shape ({shape1}) for k == 1. "
                f"{PlaquetteLayer.__name__} instances do not support empty templates."
            )
        # Check that the shape is either increasing or is constant for both
        # coordinates.
        if shape.x.slope < 0 or shape.y.slope < 0:
            raise TQECError(
                "The provided template does have a strictly decreasing shape, "
                "which will eventually lead to an empty instantiation, which is "
                f"not supported by {PlaquetteLayer.__name__} instances."
            )

    @staticmethod
    def _get_number_of_plaquettes(axis: Literal["X", "Y"], increments: int) -> int:
        # Shortening variable name for convenience
        _ew: Final[int] = EXPECTED_SPATIAL_BORDER_WIDTH
        if _ew % increments != 0:
            raise TQECError(
                f"Trying to remove {_ew} qubits from the {axis} border of a template "
                f"with increments {increments} in that axis. {_ew} % {increments} "
                "!= 0, which means that we would remove a non-integer number of "
                "plaquettes, which is not supported."
            )
        return _ew // increments

    @staticmethod
    def _get_template_shape(
        template: RectangularTemplate,
        spatial_borders_removed: frozenset[SpatialBlockBorder],
    ) -> PlaquetteScalable2D:
        """Get the shape of the provided ``template``, taking into account removed spatial borders.

        Args:
            template: template to get the shape from.
            spatial_borders_removed: a collection of all the spatial borders
                that have been removed.

        Returns:
            the shape of the provided template with the provided borders removed.

        """
        base_shape = template.scalable_shape
        # We return a shape in plaquette-coordinates. In order to know exactly the
        # number of plaquettes that will be trimmed, we need to divide
        # EXPECTED_SPATIAL_BORDER_WIDTH by the increments of the template.
        incr = template.get_increments()
        xborderp = PlaquetteLayer._get_number_of_plaquettes("X", incr.x)
        yborderp = PlaquetteLayer._get_number_of_plaquettes("X", incr.x)
        return PlaquetteScalable2D(
            base_shape.x
            - (SpatialBlockBorder.X_NEGATIVE in spatial_borders_removed) * xborderp
            - (SpatialBlockBorder.X_POSITIVE in spatial_borders_removed) * xborderp,
            base_shape.y
            - (SpatialBlockBorder.Y_NEGATIVE in spatial_borders_removed) * yborderp
            - (SpatialBlockBorder.Y_POSITIVE in spatial_borders_removed) * yborderp,
        )

    @property
    def template(self) -> RectangularTemplate:
        """Get the template stored by ``self``."""
        return self._template

    @property
    def plaquettes(self) -> Plaquettes:
        """Get the plaquettes stored by ``self``."""
        return self._plaquettes

    @property
    @override
    def scalable_shape(self) -> PhysicalQubitScalable2D:
        return self._get_scalable_shape_impl

    @cached_property
    def _get_scalable_shape_impl(self) -> PhysicalQubitScalable2D:
        tshape = PlaquetteLayer._get_template_shape(self.template, self.trimmed_spatial_borders)
        initial_qubit_offset = PhysicalQubitScalable2D(LinearFunction(0, 1), LinearFunction(0, 1))
        return tshape * self.template.get_increments() + initial_qubit_offset

    @override
    def with_spatial_borders_trimmed(self, borders: Iterable[SpatialBlockBorder]) -> PlaquetteLayer:
        # Warning: depends on the fact that plaquette indices on the border of
        # a template are ONLY on this border.
        borders = frozenset(borders)
        border_indices: set[int] = set()
        for border in borders:
            border_indices.update(self.template.get_border_indices(border.to_template_border()))
        return PlaquetteLayer(
            self.template,
            self.plaquettes.without_plaquettes(border_indices),
            trimmed_spatial_borders=self.trimmed_spatial_borders | borders,
        )

    def __eq__(self, value: object) -> bool:
        return (
            isinstance(value, PlaquetteLayer)
            and self._template == value._template
            and self._plaquettes == value._plaquettes
        )

    def __hash__(self) -> int:
        raise NotImplementedError(f"Cannot hash efficiently a {type(self).__name__}.")

    @property
    @override
    def scalable_num_moments(self) -> LinearFunction:
        return LinearFunction(
            0,
            max(
                (plaquette.num_moments for plaquette in self.plaquettes.collection.values()),
                default=0,
            ),
        )



"""ZX graph with 3D positions."""

from __future__ import annotations

from collections.abc import Mapping
from fractions import Fraction
from typing import TYPE_CHECKING

from pyzx.graph.graph_s import GraphS
from pyzx.utils import EdgeType, VertexType

from tqec.computation.block_graph import BlockGraph
from tqec.interop.pyzx.utils import cube_kind_to_zx
from tqec.utils.exceptions import TQECError
from tqec.utils.position import Direction3D, Position3D

if TYPE_CHECKING:
    from matplotlib.figure import Figure
    from mpl_toolkits.mplot3d.axes3d import Axes3D


class PositionedZX:
    def __init__(self, g: GraphS, positions: Mapping[int, Position3D]) -> None:
        """Represent a ZX graph with 3D positions and additional constraints.

        The constraints are:

        1. The vertex IDs in the graph match the position keys exactly.
        2. The neighbors are all shifted by 1 in the 3D positions.
        3. All the spiders are Z(0) or X(0) or Z(1/2) or Boundary spiders.
        4. Boundary and Z(1/2) spiders are dangling, and Z(1/2) connects to the
           time direction.
        5. There are no 3D corners.

        Args:
            g: The ZX graph.
            positions: A dictionary mapping vertex IDs to their 3D positions.

        Raises:
            TQECError: If the constraints are not satisfied.

        """
        self.check_preconditions(g, positions)

        self._g = g
        self._positions: dict[int, Position3D] = dict(positions)

    @staticmethod
    def check_preconditions(g: GraphS, positions: Mapping[int, Position3D]) -> None:
        """Check the preconditions for the ZX graph with 3D positions."""
        # 1. Check the vertex IDs in the graph match the positions
        if g.vertex_set() != set(positions.keys()):
            raise TQECError("The vertex IDs in the ZX graph and the positions do not match.")
        # 2. Check the neighbors are all shifted by 1 in the 3D positions
        for s, t in g.edge_set():
            ps, pt = positions[s], positions[t]
            if not ps.is_neighbour(pt):
                raise TQECError(
                    f"The 3D positions of the endpoints of the edge {s}--{t} "
                    f"must be neighbors, but got {ps} and {pt}."
                )
        # 3. Check all the spiders are Z(0) or X(0) or Z(1/2) or Boundary spiders
        for v in g.vertices():
            vt = g.type(v)
            phase = g.phase(v)
            if (vt, phase) not in [
                (VertexType.Z, 0),
                (VertexType.X, 0),
                (VertexType.Z, Fraction(1, 2)),
                (VertexType.BOUNDARY, 0),
            ]:
                raise TQECError(f"Unsupported vertex type and phase: {vt} and {phase}.")
            # 4. Check Boundary and Z(1/2) spiders are dangling, additionally
            # Z(1/2) connects to time direction
            if vt == VertexType.BOUNDARY or phase == Fraction(1, 2):
                if g.vertex_degree(v) != 1:
                    raise TQECError(
                        "Boundary or Z(1/2) spider must be dangling, but "
                        f"got {len(g.neighbors(v))} neighbors."
                    )
                if phase == Fraction(1, 2):
                    nb = next(iter(g.neighbors(v)))
                    vp, nbp = positions[v], positions[nb]
                    if abs(nbp.z - vp.z) != 1:
                        raise TQECError(
                            "Z(1/2) spider must connect to the time direction, "
                            f"but Z(1/2) at {vp} connects to {nbp}."
                        )
        # 5. Check there are no 3D corners
        for v in g.vertices():
            vp = positions[v]
            if len({_get_direction(vp, positions[u]) for u in g.neighbors(v)}) == 3:
                raise TQECError(f"ZX graph has a 3D corner at node {v}.")

    def __getitem__(self, v: int) -> Position3D:
        return self._positions[v]

    @property
    def g(self) -> GraphS:
        """Return the internal ZX graph."""
        return self._g

    @property
    def positions(self) -> dict[int, Position3D]:
        """Return the 3D positions of the vertices."""
        return self._positions

    @property
    def p2v(self) -> dict[Position3D, int]:
        """Return the mapping from 3D positions to vertices."""
        return {p: v for v, p in self._positions.items()}

    def get_direction(self, v1: int, v2: int) -> Direction3D:
        """Return the direction connecting two vertices."""
        p1, p2 = self[v1], self[v2]
        return _get_direction(p1, p2)

    @staticmethod
    def from_block_graph(block_graph: BlockGraph) -> PositionedZX:
        """Convert a :py:class:`.BlockGraph` to a ZX graph with 3D positions.

        The conversion process is as follows:

        1. For each cube in the block graph, convert it to a ZX vertex.
        2. For each pipe in the block graph, add an edge to the ZX graph with the corresponding
           endpoints and Hadamard flag.

        Args:
            block_graph: The block graph to be converted to a ZX graph.

        Returns:
            The :py:class:`~tqec.interop.pyzx.positioned_zx.PositionedZX` object converted from
            the block graph.

        """
        v2p: dict[int, Position3D] = {}
        p2v: dict[Position3D, int] = {}
        g = GraphS()

        for cube in sorted(block_graph.cubes, key=lambda c: c.position):
            vt, phase = cube_kind_to_zx(cube.kind)
            v = g.add_vertex(vt, phase=phase)
            v2p[v] = cube.position
            p2v[cube.position] = v

        for edge in block_graph.pipes:
            et = EdgeType.HADAMARD if edge.kind.has_hadamard else EdgeType.SIMPLE
            g.add_edge((p2v[edge.u.position], p2v[edge.v.position]), et)

        return PositionedZX(g, v2p)

    def to_block_graph(self) -> BlockGraph:  # pragma: no cover
        """Convert the positioned ZX graph to a block graph."""
        # Needs to be imported here to avoid pulling pyzx when importing this module.
        from tqec.interop.pyzx.synthesis.positioned import (  # noqa: PLC0415
            positioned_block_synthesis,
        )

        return positioned_block_synthesis(self)

    def draw(
        self,
        *,
        figsize: tuple[float, float] = (5, 6),
        title: str | None = None,
        node_size: int = 400,
        hadamard_size: int = 200,
        edge_width: int = 1,
    ) -> tuple[Figure, Axes3D]:  # pragma: no cover
        """Plot the :py:class:`~tqec.interop.pyzx.positioned.PositionedZX` using matplotlib.

        Args:
            graph: The ZX graph to plot.
            figsize: The figure size. Default is ``(5, 6)``.
            title: The title of the plot. Default to the name of the graph.
            node_size: The size of the node in the plot. Default is ``400``.
            hadamard_size: The size of the Hadamard square in the plot. Default
                is ``200``.
            edge_width: The width of the edge in the plot. Default is ``1``.

        Returns:
            A tuple of the figure and the axes.

        """
        # Needs to be imported here to avoid pulling pyzx when importing this module.
        from tqec.interop.pyzx.plot import plot_positioned_zx_graph  # noqa: PLC0415

        return plot_positioned_zx_graph(
            self,
            figsize=figsize,
            title=title,
            node_size=node_size,
            hadamard_size=hadamard_size,
            edge_width=edge_width,
        )


def _get_direction(p1: Position3D, p2: Position3D) -> Direction3D:
    """Return the direction connecting two 3D positions."""
    if p1.x != p2.x:
        return Direction3D.X
    if p1.y != p2.y:
        return Direction3D.Y
    return Direction3D.Z



from collections.abc import Sequence

from tqec.compile.observables.abstract_observable import CubeWithArms, PipeWithArms
from tqec.compile.observables.builder import Coordinates2D, ObservableBuilder
from tqec.compile.observables.fixed_bulk_builder import build_regular_cube_top_readout_qubits
from tqec.compile.specs.enums import SpatialArms
from tqec.computation.cube import ZXCube
from tqec.utils.enums import Orientation
from tqec.utils.position import Direction3D, PlaquetteShape2D, SignedDirection3D


def build_spatial_cube_top_readout_qubits(
    shape: PlaquetteShape2D, arms: SpatialArms
) -> Sequence[Coordinates2D]:
    """Build the qubit coordinates for a straight or bent middle line of the top face of a spatial
    cube.
    """
    assert len(arms) == 2
    half_x, half_y = shape.x // 2, shape.y // 2

    if arms == SpatialArms.LEFT | SpatialArms.RIGHT:
        return [(x, half_y) for x in range(1, shape.x)]
    elif arms == SpatialArms.UP | SpatialArms.DOWN:
        return [(half_x, y) for y in range(1, shape.y)]
    elif arms == SpatialArms.LEFT | SpatialArms.UP:
        return [(x, half_y) for x in range(1, half_x)] + [(half_x, y) for y in range(1, half_y)]
    elif arms == SpatialArms.DOWN | SpatialArms.RIGHT:
        return [(x, half_y) for x in range(shape.x - 1, half_x, -1)] + [
            (half_x, y) for y in range(shape.y - 1, half_y, -1)
        ]
    elif arms == SpatialArms.UP | SpatialArms.RIGHT:
        return [(x, half_y) for x in range(shape.x - 1, half_x, -1)] + [
            (half_x, y) for y in range(1, half_y + 1)
        ]
    # arms == SpatialArms.LEFT | SpatialArms.DOWN:
    return [(x, half_y) for x in range(1, half_x + 1)] + [
        (half_x, y) for y in range(shape.y - 1, half_y, -1)
    ]


def build_cube_top_readout_qubits(
    shape: PlaquetteShape2D, cube: CubeWithArms
) -> Sequence[Coordinates2D]:
    """Build the qubit coordinates whose measurements will be included in the observable on the top
    face of a cube.
    """
    if not cube.cube.is_spatial:
        kind = cube.cube.kind
        assert isinstance(kind, ZXCube)
        # Determine the middle line orientation based on the cube kind.
        # Since the basis of the top face decides the measurement basis of the data
        # qubits, i.e. the logical operator basis. We only need to find the spatial
        # boundaries that the logical operator can be attached to.
        obs_orientation = Orientation.VERTICAL if kind.y == kind.z else Orientation.HORIZONTAL
        return build_regular_cube_top_readout_qubits(shape, obs_orientation)
    return build_spatial_cube_top_readout_qubits(shape, cube.arms)


def _extended_stabilizers_used(
    pipe: PipeWithArms,
) -> bool:
    if pipe.pipe.direction == Direction3D.X:
        return False
    u_arms, v_arms = pipe.cube_arms
    return u_arms.has_spatial_arm_in_both_dimensions ^ v_arms.has_spatial_arm_in_both_dimensions


def _build_pipe_top_readout_qubits_impl(
    shape: PlaquetteShape2D, direction: Direction3D, extended_stabilizers_used: bool
) -> Sequence[Coordinates2D]:
    assert direction != Direction3D.Z
    if direction == Direction3D.X:
        return [(shape.x, shape.y // 2)]
    # Extended stabilizers at pipe do not have qubits that need to be include
    # in the observable.
    if extended_stabilizers_used:
        return []
    return [(shape.x // 2, shape.y)]


def build_pipe_top_readout_qubits(
    shape: PlaquetteShape2D, pipe: PipeWithArms
) -> Sequence[Coordinates2D]:
    """Build the qubit coordinates whose measurements will be included in the observable on the top
    face of a pipe.
    """
    return _build_pipe_top_readout_qubits_impl(
        shape, pipe.pipe.direction, _extended_stabilizers_used(pipe)
    )


def build_regular_cube_bottom_stabilizer_qubits(
    shape: PlaquetteShape2D,
    connect_to: SignedDirection3D,
) -> Sequence[Coordinates2D]:
    """Build the stabilizer measurement coordinates who will be included in the observable spanning
    half of the bottom face of a regular cube.
    """
    stabilizers: list[tuple[float, float]] = []
    # We calculate the qubits for the connect_to=SignedDirection3D(Direction3D.X, True) case
    # and rotate to get the correct orientation.
    for i in range(shape.x // 2):
        x = shape.x - i - 0.5
        for j in range(shape.y // 2):
            y = (1 - i % 2) + 2 * j + 0.5
            stabilizers.append((x, y))
    # rotate all coordinates around the block center:
    # rx = cx + a * (x - cx) - b * (y - cy)
    # ry = cy + b * (x - cx) + a * (y - cy)
    # in which (cx, cy) is the center of the block, a = cos(theta), b = sin(theta)
    # and theta is the angle of rotation.
    a, b = 1, 0
    match connect_to:
        case SignedDirection3D(Direction3D.X, False):
            a, b = -1, 0
        case SignedDirection3D(Direction3D.Y, True):
            a, b = 0, 1
        case SignedDirection3D(Direction3D.Y, False):
            a, b = 0, -1
        case _:
            pass
    cx, cy = shape.x // 2, shape.y // 2
    return [
        (
            cx + a * (x - cx) - b * (y - cy),
            cy + b * (x - cx) + a * (y - cy),
        )
        for x, y in stabilizers
    ]


def build_spatial_cube_bottom_stabilizer_qubits(
    shape: PlaquetteShape2D,
) -> Sequence[Coordinates2D]:
    """Build the stabilizer measurement coordinates who will be included in the observable spanning
    the bottom face of a single spatial cube.
    """
    return [(i + 0.5, j + 0.5) for i in range(shape.x) for j in range(shape.y) if (i + j) % 2 == 0]


def build_connected_spatial_cube_bottom_stabilizer_qubits(
    shape: PlaquetteShape2D,
    arms: SpatialArms,
    connect_to: SignedDirection3D,
    extended_stabilizers_used: bool,
) -> Sequence[Coordinates2D]:
    """Build the stabilizer measurement coordinates who will be included in the observable spanning
    the bottom face of a spatial cube connected to pipes.
    """
    max_y = (
        shape.y - 1
        if extended_stabilizers_used and connect_to == SignedDirection3D(Direction3D.Y, True)
        else shape.y
    )
    bulk_parity = 1 if arms in [SpatialArms.UP, SpatialArms.DOWN] else 0
    return [
        (i + 0.5, j + 0.5)
        for i in range(shape.x)
        for j in range(max_y)
        if (i + j) % 2 == bulk_parity
    ]


def build_pipe_bottom_stabilizer_qubits(
    shape: PlaquetteShape2D, pipe: PipeWithArms
) -> Sequence[Coordinates2D]:
    """Build the stabilizer measurement coordinates who will be included in the observable on the
    bottom face of a pipe.

    It includes the bottom stabilizers of the connected cubes.

    """
    direction = pipe.pipe.direction
    extended_stabilizers_used = _extended_stabilizers_used(pipe)

    qubits: list[Coordinates2D] = []
    for cube, arms in zip(pipe.pipe, pipe.cube_arms):
        kind = cube.kind
        assert isinstance(kind, ZXCube)
        is_u = cube == pipe.pipe.u
        connect_to = SignedDirection3D(direction, is_u)
        if cube.is_spatial:
            stabilizers = build_connected_spatial_cube_bottom_stabilizer_qubits(
                shape, arms, connect_to, extended_stabilizers_used
            )
        else:
            stabilizers = build_regular_cube_bottom_stabilizer_qubits(shape, connect_to)
        # the local coordinates is of cube u, therefore, we need to shift the coordinates
        # of the stabilizers within cube v
        if not is_u:
            stabilizers = [
                (
                    s[0] + shape.x if direction == Direction3D.X else s[0],
                    s[1] + shape.y if direction == Direction3D.Y else s[1],
                )
                for s in stabilizers
            ]
        qubits.extend(stabilizers)
    return qubits


FIXED_BOUNDARY_OBSERVABLE_BUILDER = ObservableBuilder(
    cube_top_readouts_builder=build_cube_top_readout_qubits,
    pipe_top_readouts_builder=build_pipe_top_readout_qubits,
    cube_bottom_stabilizers_builder=lambda shape, cube: build_spatial_cube_bottom_stabilizer_qubits(
        shape
    ),
    pipe_bottom_stabilizers_builder=build_pipe_bottom_stabilizer_qubits,
    pipe_temporal_hadamard_builder=lambda shape, pipe: [],
)
